/**
 * ApplicationNode.h
 *
 *  Created on: Feb 9, 2009
 *      Author: barrygr
 */

#ifndef APPLICATIONNODE_H_
#define APPLICATIONNODE_H_

#include <QtGui>
#include <QApplication>
#include <QFileDialog>
#include <H3D/X3DChildNode.h>
#include <H3D/SFInt32.h>
#include <H3D/SFString.h>
#include <H3D/DEFNodes.h>
#include <H3D/X3D.h>
#include <H3D/KeySensor.h>
#include <H3D/X3DChildNode.h>
#include <H3D/MFNode.h>

#include "RegisterUserWindow.h"
#include "SelectCaseWindow.h"
#include "RemoveWisdomToothWindow.h"
#include "GlWindowNode.h"
#include "LocalCaseList.h"
#include "ServerCaseList.h"
#include "ProgressDialog.h"
#include "DatabaseHandler.h"

#include <QDir>
#include <QFile>
#include <QFileInfo>
#include <QtNetwork/QFtp>
#include <QtNetwork/QUrlInfo>

#include <stdexcept>
#include <time.h>
#include <openssl/md5.h>
#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <iomanip>
#include <ios>
#include <sstream>
#include <vector>



#include "../src/VolumeModel.h"
#include "../src/MaterialSegmentationModel.h"
#include "../src/LoggingNode.h"
#include "../src/DrillingStepsNode.h"

using namespace H3D;

namespace FS
{

class LocalCaseList;
class ServerCaseList;
class MainWindow;
class RegisterUserWindow;
class SelectCaseWindow;
class RemoveWisdomToothWindow;
class GlWindowNode;
class ForbiddenSegmentationModel;

//old code
class Updater_QButton_SaveAndExit;

/**
 * The main class for the FsWisdom Application. Creates Windows as needed and also acts as a H3D Node
 */
class ApplicationNode : public QApplication, public H3D::X3DChildNode
{
	Q_OBJECT

public:

	/**
	 * Constructor
	 * Reads settings from settings file.
	 * Establishes an escape button handler for application exit
	 * Display the window where the user registers
	 */
	ApplicationNode(int argc, char *argv[]);

	/**
	 * Destructor
	 */
	~ApplicationNode();

	static H3D::H3DNodeDatabase database;

	/**
	 * MySQL can only handle 16 character username. Therefore we generate a hashed
	 * username from user-email to be used in mySQL.
	 */
	QString hashUsername(QString original);

	/**
	 * Registers a User. This involves creating their logging directory if it doesn't already exist.
	 * Launches the Select Case window so that the User can select which Case they wish to run.
	 */
	void RegisterUser(QString username, QString password);
	QProgressDialog dialog;
	/**
	 * Run time exception class for ApplicationNode
	 */
	class ApplicationNodeException: public runtime_error
	{
		public:

			/**
			 * Constructor
			 */
			ApplicationNodeException(const string& what) :
				runtime_error("Exception within ApplicationNode: " + what)
				{
				}
	};

	/// opens a gl window on the second screen and loads into it the scene from x3dFilePath
	/// \param full path to file
	void launchX3D(const QDir& x3dFilePath);

	/// get if the exercise is to be executed on the screen as mirrored
	bool getIsMirrored();

	/// set that the exercise is to be executed on the screen as mirrored
	void setIsMirrored(bool isMirrored);

	/// get if the exercise is to be executed on the screen as quad buffered stereo
	bool getIsStereo();

	/// set that the exercise is to be executed on the screen as quad buffered stereo
	void setIsStereo(bool isStereo);

	/// Get the username of the student
	QString getUsername();

	/// Get the logdirectory of the student
	const QDir& getLogDirectory();

	/**
	 * get the previous x3d file the was launched via the "Launch x3d file" tab
	 * \return the full file path
	 */
	const QDir& getPreviousX3dFilePath();

	const QDir& getRemoveWisdomToothX3dFilePath();

	/**
	 * set the previous x3d file the was launched via the "Launch x3d file" tab
	 * \param the full file path
	 */
	void setPreviousX3dFilePath(const QDir& previousX3dFilePath);

	void setRemoveWisdomToothX3dFilePath(const QDir& x3dFilePath);

	const QDir& getCaseDirectory() const;

	void setWaitCursor();
	void restoreNormalCursor();

	void launchX3DWithStandardGuiWindow(const QDir& x3dFilePath, bool playback, int forsCaseVersionId = -1);

	/**
	 * Get a node pointer from the current displayed scene in the gl window
	 */
	Node* getSceneNode(const string& nodeName);

	QFtp* getFtp();
	LocalCaseList* getLocalCaseList();
	ProgressDialog* getProgressDialog();
	void setFtpAborted(bool aborted);
	bool getFtpAborted();
	void closeX3dFile();


public slots:

	/// Quit the application
	void Quit()
	{
		if(ftpFinished)
			quit();
	}

	void OnAcceptUserData();

	/// Returns to the Register User Screen
	void logout();

	/**
	 * Displays a save/exit popup
	 * Data is stored in the database. Forbidden segements are stored with their segmentID negative.
	 */
	void exitX3DWithStandardGuiWindow();
	bool CleanupX3dWindows();

	void ftpCommandFinished(int commandID, bool error);
	void ftpCommandStarted(int id);
	void ftpDownloadComplete();
	void cancelFtp();
	void showRegisterUserWindow();

private:

	void InitSettings();
	bool CreateDirectoryIfNotExists( const QDir& rlogDirectoryPath );
	void RouteWidgets(bool isPlayback);

	/**
	* Retrieves the data from the users performed exercise and passes it to DatabaseHandler.
	*/
	void SavePerformedUserCase();

	/**
	 * Captures the escape so as to escape from the application
	 */
	class QuitApplicationField: public AutoUpdate< SFString >
	{
	  virtual void update()
	  {
	    string s = static_cast< SFString * >(routes_in[0])->getValue();
	    if( s[0] == 27  ) // Esc button
	    	quit();
	  }
	};

	/// Used by startExercise and startPlayback functions
	void startX3d(bool isPlayback);

	AutoRef<GlWindowNode> glWindow;				// the opengl window
	auto_ptr<RegisterUserWindow> registerUserWindow;
	auto_ptr<SelectCaseWindow> selectCaseWindow;
	auto_ptr<RemoveWisdomToothWindow> removeWisdomToothWindow;
	auto_ptr<ProgressDialog> progressDialog;

	auto_ptr< QuitApplicationField > quitApplicationField;
	AutoRef< KeySensor > keySensor;

	bool isMirrored;
	const QString isMirroredKey;
	bool isStereo;
	const QString isStereoKey;

	auto_ptr<QDir>  x3dFilePath; 			// the x3d file to run
	const QString x3dFilePathKey;

	// QT's way of saving settings until the next time the app is executed
	auto_ptr<QSettings> settings;

	QString username;
	QString password;
	auto_ptr<QDir> userLoggingDirectory;	// the users directory (within the log directory)

	int userId; // user's id in database

	int forsCaseVersionId; // the row id ov the fors case version in the database

	QDir  logDirectory; 					// the log directory
	const QString logDirectoryKey;			// the name of the log directory setting's key

	QDir  previousX3dFilePath; 				// the previous file selected under the "Launch X3d File" tab
	const QString previousX3dFilePathKey;	// the name of the previous x3d File Path setting's key

	QDir  removeWisdomToothX3dFilePath;
	const QString removeWisdomToothX3dFilePathKey;

	QDir caseDirectory;
	const QString caseDirectoryKey;

	QString databaseName;
	const QString databaseNameKey;

	QString databaseIP;
	const QString databaseIPKey;

	QString databasePassword;
	const QString databasePasswordKey;

	QString ftpUser;
	const QString ftpUserKey;

	QString ftpIP;
	const QString ftpIPKey;

	QString ftpPassword;
	const QString ftpPasswordKey;

	VolumeModel* volumeModel;
	MaterialSegmentationModel* materialSegmentationModel;
	ForbiddenSegmentationModel* forbiddenSegmentationModel;

	LoggingNode* loggingNode;
	DrillingStepsNode* drillingStepsNode;

	bool isPlayback;
    auto_ptr<LocalCaseList> localCaseList;
    auto_ptr<ServerCaseList> serverCaseList;
	auto_ptr<QFtp> ftp;
	int myCDcommand;
	bool ftpAborted;
	bool ftpFinished;
	bool loginPressedOk;
};

}
#endif /* APPLICATIONNODE_H_ */



