#ifndef DATABASEHANDLER_H_
#define DATABASEHANDLER_H_


#include <QSqlDatabase>
#include <QSqlQuery>
#include <QString>
#include <QVariant>
#include <iostream>
#include <stdexcept>
#include <vector>
#include "Case.h"

using namespace std;

/**
* Class contains the data needed to save a users case to the database.
*/
class SegmentData
{
	public:
		int SegmentId;
		int DrillingStep;
		int VoxelsRemoved;
		int noOfVoxelsEqualToExpert;
};

/**
* Class is used when a server case list is requested.
*/
class CaseMembers
{
	public :
		QString Name;
		unsigned int version;
};

/**
* A Singleton class that is responsible for reading/writing to the SQL database.
*/		
class DatabaseHandler
{
	public:
		/**
		* Destructor
		*/	 
		~DatabaseHandler();
		
		/**
		* Destroys the single instance of DatabaseHandler. Should only be called when logging out from FsWisdom
		*/	 
		void DestroyInstance();
		
		/**
		* Returns a pointer to the single instance.
		*/	 
		static DatabaseHandler* GetInstance();
		
		/**
		* Gets the Id of a selected case.
		*/	 
		unsigned int GetSelectedCaseId( const QString& rName, const unsigned int& rVersion );
		
		/**
		* Gets the Anamnes information of a selected case.
		*/	 
		bool GetAnamnesText(const QString& rCaseName, const unsigned int& rCaseVersion, QString& rAnamnesText );
		
		/**
		* Gets the objectives information of a selected case.
		*/	
		bool GetCaseObjectives( const QString& rCaseName, const unsigned int& rCaseVersion, QString& rObjectives );
		
		/**
		* Gets the server case list.
		*/	
		bool GetServerCaseList( std::vector<CaseMembers>& rCaseList );
		
		/**
		* Gets the timestamp data for a selected case and timestamp
		*/	
		bool GetTimeStampData( const QString& rSelectedCase, const QString& selectedTimestamp, QByteArray& rResult );  
		
		/**
		* Gets a list of timestamps that belongs to a selected case. 
		*/
		bool GetTimeStampList( const QString& rSelectedCase, QStringList& stringList );
		
		
		/**
		* Logs on to the database with the given credentials. 
		* If it fails to open the database or retrieve a ID for the user the function will throw an exception.  
		*/
		void RegisterUser(  const QString& rdatabaseIp, const QString& rdatabaseName, const QString& ruserName, 
					 		const QString& rhashUsername, const QString& rhashPassword );
		
		/**
		* Saves the users performed exercise to the database. 
		* The function will throw an exception if it fails to save the information.  
		*/
		void SavePerformedExercise(	const QByteArray& rBytearray,  const int& rforsCaseVersionId, 
									const std::vector<SegmentData>& rForbiddenSegmentedData,
									const std::vector<SegmentData>& rSegmentedData );
								
		
	
	protected:
		/**
		* constructor 
		*/
	    DatabaseHandler();	
		
		/**
		* copy pconstructor 
		*/
		DatabaseHandler( const DatabaseHandler& );
		
		/**
		*  Gets a Id from a user name
		*/
		int GetUserIdFromDb( const QString& ruserName );
		
		/**
		* Initializes the member variables
		*/
	    void Initialize();
	    
	    /**
		* Tries to open the database with login credentials 
		*/
	    bool OpenDbWithSavedCredentials();
	    
	    /**
		* Saves login credentials to member variables 
		*/
	    bool SaveCredentials(	const QString& rdatabaseIp, const QString& rdatabaseName, const QString& ruserName,  
					 			const QString& rhashUsername, const QString& rhashPassword );
		

		/**
		* SQL database object 
		*/
		QSqlDatabase m_forsDatabase;
	
	private:
		static DatabaseHandler *m_pDataBaseHandlerInstance;
		QString m_databaseIP;
		QString m_databaseName;
		QString m_username;
		QString m_hashUsername;
		QString m_hashPassword;			
		int m_userId; 				
		
};





/**
* Run time exception class for ApplicationNode
*/
class DatabaseHandlerException : public runtime_error
{
	public:

	/**
	* Constructor
	*/
	DatabaseHandlerException(const string& what) : runtime_error("Exception within DatabaseHandler: " + what)
	{
	}
};






#endif /* DATABASEHANDLER_H_ */