/**
 * GlWindowNode.h
 *
 *  Created on: Mar 19, 2009
 *      Author: barrygr
 */

#include "GlWindowNode.h"

using namespace FS;

// Add this node to the H3DNodeDatabase system.
H3DNodeDatabase GlWindowNode::database( "GlWindowNode",
		&(newInstance<GlWindowNode>),
        typeid( GlWindowNode ),
        &(H3DWindowNode::database) );

GlWindowNode::GlWindowNode(
	ApplicationNode *applicationNode,
	const QDir& x3dFilePath,
	Inst< SFInt32     > width,
	Inst< SFInt32     > height,
        Inst< SFBool      > fullscreen,
        Inst< SFBool      > mirrored,
        Inst< RenderMode  > _renderMode,
        Inst< SFViewpoint > viewpoint,
        Inst< SFTime      > time
        ) : H3DWindowNode( width, height, fullscreen, mirrored, _renderMode, viewpoint, time )
{
	this->applicationNode = applicationNode;

	type_name = "GlWindowNode";
	database.initFields( this );

	if(applicationNode->getIsStereo())
		setFormat(QGLFormat( QGL::DoubleBuffer  | QGL::AlphaChannel | QGL::DepthBuffer | QGL::Rgba | QGL::StencilBuffer | QGL::StereoBuffers ));
	else
		setFormat(QGLFormat( QGL::DoubleBuffer  | QGL::DepthBuffer | QGL::Rgba | QGL::AlphaChannel  | QGL::AccumBuffer | QGL::StencilBuffer  | QGL::NoStereoBuffers | QGL::DirectRendering | QGL::NoOverlay | QGL::NoSampleBuffers));

	launchX3D(x3dFilePath);
}

void GlWindowNode::launchX3D(const QDir& x3dFilePath)
{
	applicationNode->setWaitCursor();

	if(timer.get())
		timer->stop();

	// set the current working directory
	QDir currentDirectory = QDir::current();
	QDir::setCurrent(QFileInfo(x3dFilePath.path()).absolutePath());

	if(!group.get())
		group.reset( new Group );
	else
		group->children->clear();

	if(!defNodes.get())
		defNodes.reset( new X3D::DEFNodes );
	else
		defNodes->clear();

	// read from x3d file
	try
	{
		group->children->push_back( X3D::createX3DFromURL( x3dFilePath.path().toStdString(), defNodes.get() ) );

		this->mirrored->setValue(applicationNode->getIsMirrored());
		this->renderMode->setValue( applicationNode->getIsStereo() ? "QUAD_BUFFERED_STEREO" : "MONO" );

		// get a new scene
		scene.reset(new Scene);
		scene->window->push_back( this );
		scene->sceneRoot->setValue( group.get() );
		scene->idle();

		// full screen on the second monitor
		this->setGeometry(QApplication::desktop()->availableGeometry(1));
		this->showFullScreen();

		// start scene timer.
		if(!timer.get())
		{
			timer.reset(new QTimer(this));
			QObject::connect(timer.get(), SIGNAL(timeout()), this, SLOT(updateScene()));
		}

		timer->start();
	}
	catch(const H3D::X3D::XMLParseError &message)
	{
		cout << "H3D::X3D::XMLParseError in GlWindowNode: " << message << endl;
	}

	// restore the current path
	QDir::setCurrent(currentDirectory.path());
	applicationNode->restoreNormalCursor();
}

void GlWindowNode::startTimer()
{
	timer->start();
}

void GlWindowNode::stopTimer()
{
	timer->stop();
}

void GlWindowNode::release()
{
	if( scene.get() )
	{
		if( scene->window.get() )
			scene->window->clear();
	}
}

GlWindowNode::~GlWindowNode()
{
	cout << "Destroying GlWindowNode" << endl;

	ClearAll();

	cout << "Destroyed GlWindowNode" << endl;
}

void GlWindowNode::ClearAll()
{

	if( timer.get() )
		timer->stop();

	if( defNodes.get() )
		defNodes->clear();

	if( group.get() )
	{
		if( group->children.get() )
			group->children->clear();
	}
}

void GlWindowNode::DestroyOpenGLWindow()
{
	release();
	ClearAll();
}

void GlWindowNode::initWindow()
{
	glInit();
	QGLWidget::resize( H3DWindowNode::width->getValue(), H3DWindowNode::height->getValue() );
	glClearColor( 0, 0, 0, 1 );
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glClear( GL_COLOR_BUFFER_BIT );
}

void GlWindowNode::setFullscreen( bool /*fullscreen*/ )
{
}

void GlWindowNode::swapBuffers()
{
  QGLWidget::swapBuffers();
}

void GlWindowNode::makeWindowActive()
{
  QGLWidget::makeCurrent();
}

void GlWindowNode::initializeGL()
{
  QGLWidget::initializeGL();
}
void GlWindowNode::paintGL()
{
  display();
}

void GlWindowNode::resizeGL(int width, int height)
{
  QGLWidget::resizeGL( width, height );
  reshape( width, height );
  glClear( GL_COLOR_BUFFER_BIT );
}

void GlWindowNode::keyPressEvent(QKeyEvent *event)
{
	QByteArray byteArray = event->text().toAscii();

	for (int i = 0; i < byteArray.size(); i++)
		this->onKeyDown(byteArray.at(i),false);
}

void GlWindowNode::mouseMoveEvent(QMouseEvent *event)
{
	Qt::MouseButtons mouseButtons = event->buttons();
	onMouseMotionAction( event->x(), event->y() );
}

void GlWindowNode::mousePressEvent(QMouseEvent *event)
{
	Qt::MouseButtons mouseButtons = event->buttons();

	if(mouseButtons & Qt::LeftButton)
		onMouseButtonAction( H3D::MouseSensor::LEFT_BUTTON, H3D::MouseSensor::DOWN);

	if(mouseButtons & Qt::MidButton)
		onMouseButtonAction( H3D::MouseSensor::MIDDLE_BUTTON, H3D::MouseSensor::DOWN);

	if(mouseButtons & Qt::RightButton)
		onMouseButtonAction( H3D::MouseSensor::RIGHT_BUTTON, H3D::MouseSensor::DOWN);
}

void GlWindowNode::mouseReleaseEvent(QMouseEvent* /*event*/ )
{
	onMouseButtonAction( H3D::MouseSensor::LEFT_BUTTON, H3D::MouseSensor::UP);
	onMouseButtonAction( H3D::MouseSensor::MIDDLE_BUTTON, H3D::MouseSensor::UP);
	onMouseButtonAction( H3D::MouseSensor::RIGHT_BUTTON, H3D::MouseSensor::UP);
}

void GlWindowNode::wheelEvent(QWheelEvent *event)
{
	float bruteDeltaValue =  event->delta()/120.0;

	for(int i = 0; i < abs((long)bruteDeltaValue); i++  )
		if(bruteDeltaValue>0   )
			onMouseWheelAction( MouseSensor::FROM );
		else
			onMouseWheelAction( MouseSensor::TOWARDS );
}

void GlWindowNode::SetWidth(int width)
{
	(H3D::H3DWindowNode::width)->setValue(width);
}

void GlWindowNode::SetHeight(int height)
{
	(H3D::H3DWindowNode::height)->setValue(height);
}

void GlWindowNode::updateScene()
{
	for( set< Scene * >::iterator i = Scene::scenes.begin(); i != Scene::scenes.end(); i++ )
		if( (*i)->isActive() )
			(*i)->idle();
}

Node* GlWindowNode::getSceneNode(const string& nodeName)
{
	return defNodes->getNode(nodeName);
}
