#include "LocalCaseList.h"
#include <QByteArray>
using namespace std;
using namespace FS;

LocalCaseList::LocalCaseList(ApplicationNode *applicationNode, const QDir& caseDirectory, QFtp* Ftp)
	: caseDirectory(caseDirectory),
ftp(NULL),
fileFtp(NULL)
{
	this->applicationNode = applicationNode;
	// build up a list of locally installed cases
	QStringList directoryList = caseDirectory.entryList(QDir::Dirs, QDir::Name);
	for (int i = 0; i < directoryList.size(); i++)
	{
		try
		{
			Case newCase(directoryList.at(i));
			insert(begin(), std::pair<QString, Case>(newCase.getName(), newCase));
		}
		// folder found in caseDirecotry which is incorrectly formatted therefore ignore it
		catch(const Case::CaseException&){}
	}

	this->ftp = Ftp;
	connect(ftp, SIGNAL(commandFinished(int, bool)),
			this, SLOT(ftpCommandFinished(int, bool)));
}

void LocalCaseList::ftpDownload(QString fileName)
{
	cout << "LocalCaseList::ftpDownload: " << fileName.toStdString() << endl;

	if (fileName != NULL)
	{
		fileFtp = new QFile(QDir::tempPath() +  QDir::separator() + fileName);
		if (!fileFtp->open(QIODevice::WriteOnly)) {
			std::cout << "Unable to save the file" << std::endl;
			delete fileFtp;
			return;
		}
		applicationNode->getProgressDialog()->setNewText("-Welcome to FsWisdom-\n Downloading and installing cases!\n" + fileName);
		ftp->get(fileName,fileFtp);
	}
	else
	{
		emit downloadComplete();
	}
}

void LocalCaseList::ftpCommandFinished(int, bool error)
{
	if(!applicationNode->getFtpAborted())
	{
		if (ftp->currentCommand() == QFtp::ConnectToHost)
		{
			if (error)
			{
				std::cout << "CommandError ConnectToHost" << std::endl;
				return;
			}
			std::cout << "LocalCaseList CommandFinished ConnectToHost" << std::endl;
			return;
		}

		if (ftp->currentCommand() == QFtp::Login)
		{
			if (error)
			{
				return;
			}
			std::cout << "LocalCaseList CommandFinished Login" << std::endl;
			return;
		}

		if (ftp->currentCommand() == QFtp::Get) {
			if (error)
			{
				std::cout << "LocalCaseList CommandError Get" << std::endl;
				if(fileFtp)
				{
					fileFtp->close();
					fileFtp->remove();
					delete fileFtp;
				}
			}
			else
			{

				std::cout << "LocalCaseList CommandFinished Get: " << QFileInfo(*fileFtp).absoluteFilePath().toStdString() << std::endl;
				fileFtp->close();
				unzipFile(QFileInfo(*fileFtp).absoluteFilePath());
				fileFtp->remove();
				delete fileFtp;
				if (ftpDownloadVector.size() > 0)
				{
					applicationNode->getProgressDialog()->setValueDownloadingCases(ftpDownloadVector.size());
					std::cout << "LocalCaseList Vector size: " << ftpDownloadVector.size() << std::endl;
					ftpDownload(ftpDownloadVector.back());
					ftpDownloadVector.pop_back();
				}
				else
				{
					std::cout << "LocalCaseList Vector size=0" << std::endl;
					emit downloadComplete();
				}

			}
		}
	}
	else
	{
		std::cout << "File not read due to canceled ftp transfer" << std::endl;
		applicationNode->restoreNormalCursor();
		applicationNode->showRegisterUserWindow();
		applicationNode->setFtpAborted(false);
		if(fileFtp)
		{
			fileFtp->close();
			fileFtp->remove();
			delete fileFtp;
			fileFtp = NULL;
		}
		ftpDownloadVector.clear();
	}
}

void LocalCaseList::addLocally(const Case& newCase)
{
	// check if we actually have anything to install in the first place
	if(!beginCaseInstallation(newCase, false))
	{
		return;  // nothing to install
	}
	else
	{
		ftpDownloadVector.push_back(newCase.getFileArchiveName());
	}
}

void LocalCaseList::addLocally(const QDir& archiveFilePath)
{
	QFileInfo fileInfo(archiveFilePath.path());

	std::cout << "Path to Archive File: " << archiveFilePath.path().toStdString() << std::endl;

	try
	{
		Case newCase(fileInfo.baseName());

		if(!beginCaseInstallation(newCase, true))
			return;  // nothing to install

		unzipFile(archiveFilePath);
	}
	catch(const Case::CaseException&)
	{
		return;
	}
}

/*
 * Returns true if we need to install this case. Updates the map with this new case.
 * OBS: The list widget is updated and populated in SelectCaseWindow::initialize().
 * removes the old directory and creates the new directory which should then be filled.
 */
bool LocalCaseList::beginCaseInstallation(const Case& newCase, bool ignoreVersion)
{
	std::map<QString, Case>::iterator localCase = find(newCase.getName());

	if(localCase != end())
	{
		unsigned int localVersion = ((*localCase).second).getVersion();
		unsigned int newVersion = newCase.getVersion();

		// if ignoreVersion is true then we install a different version of the case otherwise
		// we install it if it is an earlier version
		if( ignoreVersion ?  (localVersion != newVersion) : (localVersion < newVersion) )
		{
			std::cout << "Another version of case previously installed" << std::endl;
			QDir directory = QDir(caseDirectory.path() + QDir::separator() + ((*localCase).second).getDirectoryName());
			std::cout << "\tTherefore removing: " << directory.path().toStdString()<<std::endl;

			if(!removeDirectory(directory))
				throw LocalCaseListException("Failed to remove previous case directory");

			erase(localCase);
		}
		else
		{
			std::cout << "Case already installed" << std::endl;
			return false;
		}
	}
	else
		std::cout << "Case not previously installed" << std::endl;

	insert(begin(), std::pair<QString, Case>(newCase.getName(), newCase));

	// create the new case directory within the cases directory
	if(!caseDirectory.mkdir(caseDirectory.path() + QDir::separator() + newCase.getDirectoryName()))
		throw LocalCaseListException("Failed to create new case directory");

	return true;
}

void LocalCaseList::unzipFile(const QDir& zipFilePath)
{
	std::cout << "Start unzip of: " << zipFilePath.path().toStdString() << std::endl;
	QuaZip zip(zipFilePath.path());
	if(!zip.open(QuaZip::mdUnzip))
		throw LocalCaseListException("Failed to open file archive");  // zip.getZipError()

	zip.setFileNameCodec("utf8");

	QuaZipFile file(&zip);
	QString fileName=NULL;

	for(bool more=zip.goToFirstFile(); more; more=zip.goToNextFile())
	{
		fileName=file.getActualFileName();

		if(!fileName.endsWith("/"))	// if not a directory
		{
			file.open(QIODevice::ReadOnly);

			if(file.getZipError()!=UNZ_OK)
				throw LocalCaseListException("Unable to open structure within file archive");  // zip.getZipError()

			QFile out(caseDirectory.path() + QDir::separator() + QFileInfo(zipFilePath.path()).baseName() + QDir::separator() +fileName);
			out.open(QIODevice::WriteOnly);

			// Read whole unzipped file before writing to output file
			QByteArray array = file.readAll();
			out.write(array);

			out.close();
			if(file.getZipError()!=UNZ_OK)
				  throw LocalCaseListException("Unable to read structure within file archive");  // zip.getZipError()

			if(!file.atEnd())
				throw LocalCaseListException("Didn't read to end of structure within file archive");

			file.close();
			if(file.getZipError()!=UNZ_OK)
				throw LocalCaseListException("Unable to close structure within file archive");
		}
		else
			if(!caseDirectory.mkdir(caseDirectory.path() + QDir::separator() + QFileInfo(zipFilePath.path()).baseName() + QDir::separator()+fileName))
				throw LocalCaseListException("Unable to create direcotry within new case directory");
	}

	zip.close();

	if(zip.getZipError()!=UNZ_OK)
		throw LocalCaseListException("Unable to close file archive");
}

/**
 * Returns true if successful
 */
bool LocalCaseList::removeDirectory(QDir &directory)
{
	bool isSuccessful = true;

	if (directory.exists())
	{
		QFileInfoList entries = directory.entryInfoList(QDir::NoDotAndDotDot | QDir::Hidden | QDir::Dirs | QDir::Files);
		int count = entries.size();

		for (int i = 0; ((i < count) && (isSuccessful == true)); i++)
		{
			QFileInfo entryInfo = entries[i];
			QString path = entryInfo.absoluteFilePath();
			if (entryInfo.isDir())
			{
				QDir innerDirectory = QDir(path);
				isSuccessful = removeDirectory(innerDirectory);
			}
			else
			{
				QFile file(path);
				if (!file.remove())
					isSuccessful = false;
			}
		}
		if (!directory.rmdir(directory.absolutePath()))
			isSuccessful = false;
	}

	return(isSuccessful);
}

bool LocalCaseList::getX3dFilePathForSelectedRow(QDir& x3dFilePathForSelectedRow, QListWidget* const listWidget)
{
	int currentlySelectedRowIndex = listWidget->currentRow();

	if(currentlySelectedRowIndex>=0)
	{
		QString selectedCaseDirectoryName = listWidget->item(currentlySelectedRowIndex)->text();
		x3dFilePathForSelectedRow = QDir(caseDirectory.path() + QDir::separator() + selectedCaseDirectoryName + QDir::separator() + "x3d" + QDir::separator() + "index.x3d");

		return true;
	}
	return false;
}

const Case LocalCaseList::getCaseForSelectedRow(QListWidget* const listWidget)
{
	int currentlySelectedRowIndex = listWidget->currentRow();
	std::cout << "getCaseForSelectedRow: " << listWidget->item(currentlySelectedRowIndex)->text().toStdString() << std::endl;
	return Case(listWidget->item(currentlySelectedRowIndex)->text());
}

/**
 * returns a string list with the directory name of each case stored herein
 * e.g. "Remove Wisdom Tooth v1"
 */
QStringList LocalCaseList::getStringList() const
{
	QStringList stringList;

	const_iterator localCase = begin();
	for(localCase=begin(); localCase!=end(); localCase++)
		stringList << ((*localCase).second).getDirectoryName();

	return stringList;
}
