#include "SelectCaseWindow.h"

using namespace FS;

SelectCaseWindow::SelectCaseWindow(ApplicationNode *applicationNode, QWidget *parent)
: QMainWindow(parent)
{
	this->applicationNode = applicationNode;
	ui.setupUi(this);

	if(applicationNode->getIsMirrored())
		ui.isMirroredCheckBox->setCheckState(Qt::Checked);
	else
		ui.isMirroredCheckBox->setCheckState(Qt::Unchecked );

	if(applicationNode->getIsStereo())
		ui.isStereoCheckBox->setCheckState(Qt::Checked);
	else
		ui.isStereoCheckBox->setCheckState(Qt::Unchecked );

	if(!applicationNode->getPreviousX3dFilePath().path().isEmpty())
	{
		ui.launchX3dLineEdit->setText(applicationNode->getPreviousX3dFilePath().path());
		ui.launchX3dFilePushButton->setEnabled(true);
	}
	else
		ui.launchX3dFilePushButton->setEnabled(false);

	QObject::connect(ui.logoutPushButton, SIGNAL(clicked ()), applicationNode, SLOT(logout()));

	QObject::connect(ui.isMirroredCheckBox, SIGNAL(stateChanged (int)), this, SLOT(setIsMirrored(int)));
	QObject::connect(ui.isStereoCheckBox, SIGNAL(stateChanged (int)), this, SLOT(setIsStereo(int)));

	QObject::connect(ui.browseToX3dFilePushButton, SIGNAL(clicked ()), this, SLOT(browseToX3dFile()));
	QObject::connect(ui.launchX3dFilePushButton, SIGNAL(clicked ()), this, SLOT(launchX3dFile()));
	QObject::connect(ui.closeX3dFilePushButton, SIGNAL(clicked ()), this, SLOT(closeX3dFile()));

	QObject::connect(ui.addCasePushButton, SIGNAL(clicked ()), this, SLOT(browseToCaseFile()));
	QObject::connect(ui.launchCaseWithPlaybackPushButton, SIGNAL(clicked ()), this, SLOT(launchCaseWithPlayback()));
	QObject::connect(ui.launchCaseWithoutPlaybackPushButton, SIGNAL(clicked ()), this, SLOT(launchCaseWithoutPlayback()));


	QObject::connect(ui.caseListWidget, SIGNAL(itemSelectionChanged ()), this, SLOT(updateCaseListTimestampsWidget()));
	QObject::connect(ui.caseListWidget, SIGNAL(itemSelectionChanged ()), this, SLOT(updateAnamnesTextWidget()));
	QObject::connect(ui.caseListWidget, SIGNAL(itemSelectionChanged ()), this, SLOT(updateObjectivesTextWidget()));
	QObject::connect(ui.caseListWidget, SIGNAL(itemSelectionChanged ()), this, SLOT(updateImageViewer()));

	//standardOutputToTextEdit.reset(new QDebugStream(std::cout, ui.standardOutputPlainTextEdit,QDir("stdout.txt")));
	//standardErrorToTextEdit.reset(new QDebugStream(std::cerr, ui.standardErrorPlainTextEdit,QDir("stderr.txt")));

	initialize();

}

void SelectCaseWindow::initialize()
{
	//Downloading of usecases is already done. Populate list-widget.
	ui.caseListWidget->clear();
	ui.label_7->setFocus();
	ui.objectivesTextWidget->clear();
	ui.anamnesTextWidget->clear();
	ui.caseListTimestampsWidget->clear();
	ui.selectCasePictureLabel->setPixmap(QPixmap(QString::fromUtf8(":/img/forsslundSystemsLogoSmall.png")));
	ui.caseListWidget->setSortingEnabled(true);
	map<QString, Case>::iterator it = applicationNode->getLocalCaseList()->begin();
	for(it=applicationNode->getLocalCaseList()->begin(); it!=applicationNode->getLocalCaseList()->end(); it++)
	{
		cout << "SelectCaseWindow::initialize(): "<< ((*it).second).getDirectoryName().toStdString() << endl;
		ui.caseListWidget->addItem(((*it).second).getDirectoryName());
	}


}

SelectCaseWindow::~SelectCaseWindow()
{
	cout << "Destroying SelectCaseWindow"<<endl;
}

void SelectCaseWindow::showFullScreenAndUpdate()
{
	/*Update timestamp list for selected case, i.e. the case that user exited from,
	  in case user has saved a new timestamp.*/
	updateCaseListTimestampsWidget();
	showFullScreen();
}

void SelectCaseWindow::setIsMirrored(int state)
{
	applicationNode->setIsMirrored(state==Qt::Checked);
}

void SelectCaseWindow::setIsStereo(int state)
{
	applicationNode->setIsStereo(state==Qt::Checked);
}

void SelectCaseWindow::browseToX3dFile()
{
	QString fileName;
	QFileDialog fileDlg(this, "Browse to X3D File",applicationNode->getPreviousX3dFilePath().absolutePath(),"X3D Files (*.x3d)");
	if(fileDlg.exec() == QDialog::Accepted) // only store new path if user pressed Ok.
	{
		// Only one file can be selected, Dlg keeps result in QStringList though
		fileName = fileDlg.selectedFiles().first();

		ui.launchX3dLineEdit->setText(fileName);
		applicationNode->setPreviousX3dFilePath(fileName);
		ui.launchX3dFilePushButton->setEnabled(true);

		cout << fileName.toStdString() << endl;
	}
}

void SelectCaseWindow::launchX3dFile()
{
	if(!applicationNode->getPreviousX3dFilePath().path().isEmpty())
	{
		ui.standardOutputPlainTextEdit->clear();
		ui.standardErrorPlainTextEdit->clear();

		applicationNode->launchX3D(applicationNode->getPreviousX3dFilePath());
		setCloseX3DFileEnable(true);
	}
}

void SelectCaseWindow::closeX3dFile()
{
	applicationNode->closeX3dFile();
	setCloseX3DFileEnable(false);
}

void SelectCaseWindow::setCloseX3DFileEnable(bool enable)
{
	ui.closeX3dFilePushButton->setEnabled(enable);
}

void SelectCaseWindow::browseToCaseFile()
{
	QString fileName = QFileDialog::getOpenFileName(this, tr("Browse to Case File"), "/home/", tr("Fors Files (*.fors)"));

	applicationNode->setWaitCursor();
	applicationNode->processEvents();

	// if a file is selected
//	if(!fileName.isEmpty())
	//	localCaseList->addLocally(QDir(fileName));

	applicationNode->restoreNormalCursor();
}

void SelectCaseWindow::launchCaseWithoutPlayback()
{
	launchCase(false);
}

void SelectCaseWindow::launchCaseWithPlayback()
{
	launchCase(true);
}

void SelectCaseWindow::launchCase(bool isPlayback)
{
	QDir x3dFilePath;
	if(applicationNode->getLocalCaseList()->getX3dFilePathForSelectedRow(x3dFilePath, ui.caseListWidget))
	{
		int forsCaseVersionId = -1;
		int currentlySelectedRowIndex = -1;
		const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);

		QString caseName = selectedCase.getName();
		unsigned int caseVersion = selectedCase.getVersion();
		forsCaseVersionId = DatabaseHandler::GetInstance()->GetSelectedCaseId( caseName, caseVersion );

		if (isPlayback)
		{
			//Playback button has been pressed
			this->getCurrentTimestampForPlayback();
			currentlySelectedRowIndex = ui.caseListTimestampsWidget->currentRow();
		}
		if(isPlayback && currentlySelectedRowIndex<0){
			//If no playback file has been selected, display a warning
			QMessageBox::warning( this, "FsWisdom",
			                     "\nNo performed exercise selected! \n");
		}
		else{
			//Check if x3d file exists
			QFileInfo fi( x3dFilePath.path());

			if( !fi.exists() )
			{
				cout<< "Missing file " << x3dFilePath.path().toStdString() << "aborting launch" <<endl;
				QMessageBox::critical( this, "FsWisdom", "\nYour cases has been deleted! Please log out and log in again to reload cases.\n");
			}
			else
			{
				//Display application window
				cout<< " the File to be lauched now " << x3dFilePath.path().toStdString()<<endl;
				applicationNode->launchX3DWithStandardGuiWindow(x3dFilePath, isPlayback, forsCaseVersionId);
			}
		}


		ui.label_7->setFocus(); // remove focus from widgets
	}
	else{
		//If no case has been selected, display a warning
					QMessageBox::warning( this, "FsWisdom",
					                     "\nNo case selected! \n");

	}
}


void SelectCaseWindow::getCurrentTimestampForPlayback()
{
	//Download currentTimestamp
	cout<<"Downloading the selected timestampfile"<<endl;

	QStringList stringList;

	int currentlySelectedRowIndex = ui.caseListTimestampsWidget->currentRow();

	if(currentlySelectedRowIndex>=0)
	{
		QString selectedTimestamp = ui.caseListTimestampsWidget->item(currentlySelectedRowIndex)->text();
		const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);
		QByteArray byteArray;

		this->applicationNode->setOverrideCursor(QCursor(Qt::WaitCursor));

		if( DatabaseHandler::GetInstance()->GetTimeStampData( selectedCase.getName(), selectedTimestamp, byteArray ) )
		{
			QString timestampPath = applicationNode->getLogDirectory().path()+"/"+applicationNode->getUsername();

			QFile file(timestampPath +  QDir::separator() + "timeStamp.nrrd");
			if (file.open(QIODevice::WriteOnly))
			{
				file.write(byteArray);
				file.close();
				std::cout << "File downloaded: " << QFileInfo(file).absoluteFilePath().toStdString() << std::endl;
			}
			else
				std::cout << "Failed to write to file: " << QFileInfo(file).absoluteFilePath().toStdString() << std::endl;
		}
		else
		{
			std::cout << "Failed to read archive file from server: " << std::endl;
		}
		this->applicationNode->restoreOverrideCursor();

	}
}



void SelectCaseWindow::updateCaseListTimestampsWidget()
{
	QStringList stringList;

	const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);

	if( DatabaseHandler::GetInstance()->GetTimeStampList( selectedCase.getName(), stringList ) )
	{
		// display a sorted list of cases in the listWidget
		ui.caseListTimestampsWidget->clear();
		ui.caseListTimestampsWidget->addItems(stringList);
		ui.caseListTimestampsWidget->setSortingEnabled(true);
	}
	else
	{
		std::cout << "Failed to get TimeStampList" << std::endl;
	}
}

void SelectCaseWindow::updateAnamnesTextWidget()
{

	const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);
	QString sAnamnesText;

	if( DatabaseHandler::GetInstance()->GetAnamnesText( selectedCase.getName(), selectedCase.getVersion(), sAnamnesText ) )
	{
		// display a sorted list of cases in the listWidget
		ui.anamnesTextWidget->clear();
		ui.anamnesTextWidget->setPlainText( sAnamnesText );
	}
	else
	{
		std::cout << "Failed to get anamnes text" << std::endl;
	}
}

void SelectCaseWindow::updateObjectivesTextWidget()
{

		const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);
		QString sObjectives;

		if( DatabaseHandler::GetInstance()->GetCaseObjectives( selectedCase.getName(), selectedCase.getVersion(), sObjectives ) )
		{
			// display a sorted list of cases in the listWidget
			ui.objectivesTextWidget->clear();
			ui.objectivesTextWidget->setPlainText( sObjectives );
		}
		else
		{
			std::cout << "Failed to get objectives for case" << std::endl;
		}
}


void SelectCaseWindow::updateImageViewer()
{
	const Case& selectedCase = applicationNode->getLocalCaseList()->getCaseForSelectedRow(ui.caseListWidget);
	QString fileName = applicationNode->getCaseDirectory().path()+"/"+selectedCase.getDirectoryName() + "/pictures/selectCasePicture.jpg";

    if (!fileName.isEmpty()) {
        QImage image(fileName);
        if (image.isNull()) {
            QMessageBox::information(this, tr("Image Viewer"),
                                     tr("Cannot load %1.").arg(fileName));
            return;
        }
        ui.selectCasePictureLabel->setPixmap(QPixmap::fromImage(image));

//      	ui.selectCasePictureLabel->adjustSize();
    }
}

