/**
 * MaterialSegmentationModelTest.h
 *
 *  Created on: Feb 3, 2009
 *      Author: barrygr
 */

#ifndef MATERIALSEGMENTATIONMODELTEST_H_
#define MATERIALSEGMENTATIONMODELTEST_H_

#include <cppunit/TestFixture.h>
#include <cppunit/extensions/HelperMacros.h>
#include "../src/MaterialSegmentationModel.h"
#include <H3D/X3D.h>

using namespace FS;
using namespace H3D;
using namespace X3D;

class MaterialSegmentationModelTest : public CppUnit::TestFixture
{
	CPPUNIT_TEST_SUITE( MaterialSegmentationModelTest );
	CPPUNIT_TEST( test );
	CPPUNIT_TEST_SUITE_END();
private:

public:
  void setUp()
  {
  }

  void tearDown()
  {
  }

  void test()
  {
	  string xmlString;

	  // test that exception is thrown if segmentIdField is missing
	  xmlString = "<MaterialSegmentationModel DEF='SM'/>";
	  CPPUNIT_ASSERT_THROW(H3D::X3D::createX3DNodeFromString(xmlString), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that exception is thrown if AIR_SEGMENT_ID is used for another segment
	  stringstream xmlStream1;
	  xmlStream1 << "<MaterialSegmentationModel DEF='SM'" <<
		  " segmentIdField = '" << (int)AIR_SEGMENT_ID << "'" <<
		  "></MaterialSegmentationModel>";
	  CPPUNIT_ASSERT_THROW(H3D::X3D::createX3DNodeFromStream(xmlStream1), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that exception is thrown if segmentNameField length is not the same as segmentIdField
	  xmlString = "<MaterialSegmentationModel DEF='SM'"
		  " segmentIdField = '64'"
		  "></MaterialSegmentationModel>";
	  CPPUNIT_ASSERT_THROW(H3D::X3D::createX3DNodeFromString(xmlString), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that exception is thrown if AIR_SEGMENT_NAME is used for another segment
	  stringstream xmlStream2;
	  xmlStream2 << "<MaterialSegmentationModel DEF='SM'" <<
	  	  " segmentIdField = '64'" <<
		  " segmentNameField = '" << AIR_SEGMENT_NAME << "'" <<
		  "></MaterialSegmentationModel>";
	  CPPUNIT_ASSERT_THROW(H3D::X3D::createX3DNodeFromStream(xmlStream2), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that exception is thrown if segmentHardnessField length is not the same as segmentIdField
	  xmlString = "<MaterialSegmentationModel DEF='SM'"
		  " segmentIdField = '64'"
		  " segmentNameField = 'Mandibular Nerve'>"
		  "</MaterialSegmentationModel>";
	  CPPUNIT_ASSERT_THROW(H3D::X3D::createX3DNodeFromString(xmlString), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that no exception is thrown for valid data
	  xmlString = "<MaterialSegmentationModel DEF='SM'"
		  " segmentIdField = '64'"
		  " segmentNameField = 'Mandibular Nerve'"
		  " segmentHardnessField = '1.2'"
		  "></MaterialSegmentationModel>";
	  CPPUNIT_ASSERT_NO_THROW(H3D::X3D::createX3DNodeFromString(xmlString));

	  // test that GetNoOfSegments returns the value 5, since air is also a segment
	  xmlString = xmlString = "<MaterialSegmentationModel DEF='SM'"
		  " segmentIdField = '64 128 192 255'"
		  " segmentNameField ='\"Mandibular Nerve\" \"Bone\" \"Tooth bone\" \"Tooth enamel\"'"
		  " segmentHardnessField = '1.9 1.05 0.03 0.02'>"
		  "</MaterialSegmentationModel>";

	  AutoRef< Node > node = H3D::X3D::createX3DNodeFromString(xmlString);
	  MaterialSegmentationModel* materialSegmentationModel = dynamic_cast<MaterialSegmentationModel*>(node.get());
	  CPPUNIT_ASSERT(materialSegmentationModel->getNoOfSegments() == 5);


	  // test that we get the name of air
	  CPPUNIT_ASSERT(AIR_SEGMENT_NAME == materialSegmentationModel->getName(AIR_SEGMENT_ID));

	  // test that we get the name of segment with id 64
	  CPPUNIT_ASSERT("Mandibular Nerve" == materialSegmentationModel->getName(64));

	  // test that we get an expection when we have an id that does not exist
	  CPPUNIT_ASSERT_THROW(materialSegmentationModel->getName(65), MaterialSegmentationModel::MaterialSegmentationModelException);


	  // test that we get air's hardness
	  CPPUNIT_ASSERT(AIR_SEGMENT_HARDNESS == materialSegmentationModel->getHardness(AIR_SEGMENT_ID));

	  // test that we get the hardness of segment with id other than that of air
	  CPPUNIT_ASSERT(fabs((float)1.9-materialSegmentationModel->getHardness(64))<numeric_limits<float>::epsilon());

	  // test that we get an expection when we have an id that does not exist
	  CPPUNIT_ASSERT_THROW(materialSegmentationModel->getHardness(65), MaterialSegmentationModel::MaterialSegmentationModelException);


	  // test that we get an expection when we try to set the hardness for air
	  CPPUNIT_ASSERT_THROW(materialSegmentationModel->setHardness(AIR_SEGMENT_ID,34.1), MaterialSegmentationModel::MaterialSegmentationModelException);

	  // test that we set the hardness of segment with id other than that of air
	  materialSegmentationModel->setHardness(64,9.9);
	  CPPUNIT_ASSERT(fabs((float)9.9-materialSegmentationModel->getHardness(64))<numeric_limits<float>::epsilon());

	  // test that we get an expection when we try to set the hardness for a non existing segmentId
	  CPPUNIT_ASSERT_THROW(materialSegmentationModel->setHardness(65,9.9), MaterialSegmentationModel::MaterialSegmentationModelException);


	  // insure that we get a valid list of segment ids
	  const std::vector<unsigned char> segmentIds = materialSegmentationModel->getSegmentIds();
	  CPPUNIT_ASSERT(segmentIds.size() == 5);
	  CPPUNIT_ASSERT(segmentIds[0] == AIR_SEGMENT_ID);
	  CPPUNIT_ASSERT(segmentIds[1] == 64);
	  CPPUNIT_ASSERT(segmentIds[4] == 255);
  }

};


#endif /* MATERIALSEGMENTATIONMODELTEST_H_ */
