//////////////////////////////////////////////////////////////////////////////
//    Copyright 2008 Forsslund Systems AB
//
//    ADrillableNode is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    ADrillableNode is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ADrillableNode; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    If you wish to use the code in a proprietary project, please contact
//    jonas@forsslundsystems.se for more information.
//
//////////////////////////////////////////////////////////////////////////////
#include "ADrillableNode.h"

using namespace FS;
using namespace std;

H3D::H3DNodeDatabase ADrillableNode::database( "ADrillableNode",
		&(newInstance<ADrillableNode>),
		typeid( ADrillableNode ),
		&X3DChildNode::database );

namespace ADrillableNodeInternals
{
H3D::FIELDDB_ELEMENT( ADrillableNode, position, INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT( ADrillableNode, cutRadius,  INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT( ADrillableNode, drillSpeed,  INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT( ADrillableNode, volumeModel, INPUT_OUTPUT);
H3D::FIELDDB_ELEMENT( ADrillableNode, materialSegmentationModel, INPUT_OUTPUT);
H3D::FIELDDB_ELEMENT( ADrillableNode, volumeDistribution, INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT( ADrillableNode, drillingStepsField, INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT( ADrillableNode, currentFPS, INPUT_OUTPUT ) //this should only be OUTPUT but doesn't seem to work...?
H3D::FIELDDB_ELEMENT( ADrillableNode, hasPedal, INPUT_OUTPUT )
}


ADrillableNode::ADrillableNode(H3D::Inst< H3D::SFVec3f > _position,
		H3D::Inst< H3D::SFFloat > _cutRadius,
		H3D::Inst< H3D::SFFloat > _drillSpeed,
		H3D::Inst< H3D::SFNode  > _volumeModel,
		H3D::Inst< H3D::SFNode  > _segmentModel,
		H3D::Inst< H3D::MFFloat >_volumeDistribution,
		H3D::Inst< H3D::SFNode  > drillingStepsField,
		H3D::Inst< H3D::SFFloat  > currentFPS,
		H3D::Inst< H3D::SFBool  > hasPedal
):
	position( _position ),
	cutRadius( _cutRadius ),
	drillSpeed( _drillSpeed ),
	volumeModel( _volumeModel ),
	materialSegmentationModel( _segmentModel ),
	volumeDistribution(_volumeDistribution),
	drillingStepsField(drillingStepsField),
	currentFPS(currentFPS),
	hasPedal(hasPedal)
	{


	type_name = "ADrillableNode";
	database.initFields( this );
	/* Default values */
	position->setValue(H3D::Vec3f(0, 0, 0));
	cutRadius->setValue(3);
	drillSpeed->setValue(1);
	updateInterval = 0.1;
	}

void ADrillableNode::initialize()
{
	cout<<"Initializing Drillable Node"<<endl;

	volume_model = static_cast< VolumeModel * >( volumeModel->getValue() );
	drillingStepsNode = static_cast< DrillingStepsNode * >( drillingStepsField->getValue() );

	if(!volume_model)
	{
		cout << "Null pointer exception in volume model!\n";
		return;
	}

	frameCount = 0;
	ptime = H3D::TimeStamp();
	firstVoxel = true;
	hasPedal->setValue(false);

	X3DChildNode::initialize();

	cout<<"Initialized Drillable Node"<<endl;
}


void ADrillableNode::traverseSG( H3D::TraverseInfo &ti )
{
	//Get current time
	double currentTime = H3D::TimeStamp();
	//For FPS calculation
	frameCount += 1;
	double deltaTime = currentTime - ptime;

	//Time since the last time a drill was made
	double diffTime = currentTime-prevTime;

	//If drilling is enabled
	if( ti.hapticsEnabled() && drillSpeed->getValue() > 0)
	{
		//Due to the low resolution in the segmented volume a minimum amount of drilling is set.
		//If the amount to be drilled is less, then the algorithm waits another turn.
		if (diffTime > updateInterval)
		{
			cutVolume(ti.getAccInverseMatrix()*position->getValue(),drillSpeed->getValue()*diffTime);
			prevTime= H3D::TimeStamp();
		}
	}

	else
	{
		prevTime = H3D::TimeStamp();
	}
	//Calculate current FPS. This is displayed in GUI.
	if (deltaTime > 1.0)
	{
		currentFPS->setValue(frameCount/deltaTime);
	    ptime = H3D::TimeStamp();
	    frameCount=0;
	}
}

/// Cuts the volume
/**
 *  \param center the center of the volume to cut
 *	\param drillImpact drillspeed times the time it has been applied
 */
void ADrillableNode::cutVolume(H3D::Vec3f center, float drillImpact)
{
	volume_model = static_cast< VolumeModel * >( volumeModel->getValue() );
	segment_model = static_cast< MaterialSegmentationModel * >( materialSegmentationModel->getValue() );

	double elapsedTime = volume_model->getElapsedTime();

	if(!volume_model)
	{
		//TODO: Replace by throw
		cout << "Null pointer exception in volume model!\n";
		return;
	}

	float radius = cutRadius->getValue();
	float distanceSquareX, distanceSquareY, distanceSquareZ;
	bool xChanged, yChanged;
	float radiusSquare = radius * radius;

	//Get the local coordinates of the maximum and minimum values of the bounding box
	Vec3i aMax = volume_model->getIndexLocal(H3D::Vec3f(center.x+radius,center.y+radius,center.z+radius));
	Vec3i aMin = volume_model->getIndexLocal(H3D::Vec3f(center.x-radius,center.y-radius,center.z-radius));

	float voxelsum=0;
	if (aMax.x!=-1 || aMax.y!=-1 || aMax.z!=-1 || aMin.x!=-1 || aMin.y!=-1 || aMin.z!=-1)
	{
		Vec3i at;
		Vec3i max = volume_model->getIndexLocalWall(H3D::Vec3f(center.x+radius,center.y+radius,center.z+radius));
		Vec3i min = volume_model->getIndexLocalWall(H3D::Vec3f(center.x-radius,center.y-radius,center.z-radius));

		for (at.x=min.x;at.x<=max.x;at.x++)
		{
			xChanged = true;
			for (at.y=min.y;at.y<=max.y;at.y++)
			{
				yChanged = true;
				for (at.z=min.z;at.z<=max.z;at.z++)
				{
					H3D::Vec3f gPos = volume_model->getGlobalCoordinate(at);
					H3D::Vec3f dist = gPos - center;
					if(xChanged)
					{
						distanceSquareX = dist.x * dist.x;
						xChanged = false;
					}
					if(yChanged)
					{
						distanceSquareY = dist.y * dist.y;
						yChanged = false;
					}
					//voxel is within the drill radius
					if ((distanceSquareX + distanceSquareY + dist.z*dist.z) <= radiusSquare)
					{

						//Voxel has some material remaining (0=air)
						float val = volume_model->getMaterialRemaining(at);
						int segmentID = volume_model->getMaterialSegmentationId(at);

						if (val == 0 )
						{
							//There is no material remaining i.e consider it as air in the distribution
							volumeDistribution->setValue(0,volumeDistribution->getValue()[0] + 1);
							voxelsum++;
						}
						else
						{
							//Add to the distribution of segments
							/*TODO: This line crash on windows when segmentID is 4, i.e. reference outside the vector
							since volumeDistribution is set to length 4 in .x3d file. On linux this is a classic example of
							a very hard bug to find - memory "overwrite".*/
							//volumeDistribution->setValue(segmentID,volumeDistribution->getValue()[segmentID] + 1);
							voxelsum++;

							//Calculate new value of material remaining
							float newValue = val - drillImpact * segment_model->getHardness(segmentID);

							volume_model->setMaterialRemaining(at,newValue);

							if (volume_model->getMaterialRemaining(at) < 0.005)
							{
								if (firstVoxel)
								{ //Start the timer
									volume_model->resetStartTime();
									firstVoxel = false;
									//TODO: Will this mean that elapsedTime will be 0 for first removed voxel?
									//TODO: Is that ok or maybe resetStartTime should do startTime = currentTime-1?
									elapsedTime = volume_model->getElapsedTime();
								}
								volume_model->setMaterialRemaining(at,0.0);
								volume_model->setImageValue(at,0.0,0.0,0.0);
								volume_model->setTimeStamp(at,elapsedTime);
								volume_model->incrementNoOfVoxelsBoredByUserForForbiddenSegment(at);
								volume_model->incrementNoOfVoxelsBoredByUserForSegment(at);
								drillingStepsNode->setAsBored(at,elapsedTime);

							}
						}
					}
				}
			}
		}
	}

	if(voxelsum>0)
	{
		for ( int i =0;i<(int)volumeDistribution->getSize(); i++ )
			volumeDistribution->setValue(i,volumeDistribution->getValue()[i]/((float)voxelsum));
	}

}


