/**
 * ExpertNode.cpp
 *
 *  Created on: Jan 27, 2009
 *      Author: barrygr
 */

#include "DrillingStepsNode.h"

using namespace FS;

// Add this node to the H3DNodeDatabase system.
H3D::H3DNodeDatabase DrillingStepsNode::database(
		"DrillingStepsNode",
		&(newInstance<DrillingStepsNode>),
		typeid(DrillingStepsNode),
		&X3DChildNode::database);

namespace DrillingStepsNodeInternals
{
H3D::FIELDDB_ELEMENT(DrillingStepsNode, engageHighlightField, OUTPUT_ONLY );
H3D::FIELDDB_ELEMENT(DrillingStepsNode, volumeModelField, INITIALIZE_ONLY);
H3D::FIELDDB_ELEMENT(DrillingStepsNode, materialSegmentationModelField,	INITIALIZE_ONLY);
H3D::FIELDDB_ELEMENT(DrillingStepsNode, forbiddenSegmentationModelField,INITIALIZE_ONLY);
}

DrillingStepsNode::DrillingStepsNode(H3D::Inst< H3D::SFBool > engageHighlightField,
		H3D::Inst< H3D::SFNode > volumeModelField,
		H3D::Inst< H3D::SFNode > materialSegmentationModelField,
		H3D::Inst< H3D::SFNode > forbiddenSegmentationModelField
)
:engageHighlightField(engageHighlightField),
volumeModelField(volumeModelField),
materialSegmentationModelField(materialSegmentationModelField),
forbiddenSegmentationModelField(forbiddenSegmentationModelField)
{
	cout << "Drilling Steps Node Constructing"<<endl;

	type_name = "DrillingStepsNode";
	database.initFields( this );

	cout << "Drilling Steps Node Constructed"<<endl<<endl;
}

void DrillingStepsNode::initialize()
{
	cout<<"Initializing Drilling Steps Node"<<endl;

	volumeModelNode = static_cast< VolumeModel * >( volumeModelField->getValue() );
	materialSegmentationModelNode = static_cast< MaterialSegmentationModel * >( materialSegmentationModelField->getValue() );
	forbiddenSegmentationModelNode = static_cast< ForbiddenSegmentationModel * >( forbiddenSegmentationModelField->getValue() );

	// find all the unique step ids within expertDrillingSteps excluding NOT_DRILLED_BY_EXPERT_STEP_ID
	std::vector<unsigned char> uniqueExpertDrillingStepsIds;

	// loop through all voxels
	for(unsigned int i =0;i<volumeModelNode->getWidth();i++)
		for(unsigned int j =0;j<volumeModelNode->getHeight();j++)
			for(unsigned int k =0;k<volumeModelNode->getDepth();k++)
			{
				Vec3i vec(i,j,k);
				unsigned char expertDrillingStepId = volumeModelNode->getExpertDrillingStepId(vec);

				// ignore non drilled voxels
				if(expertDrillingStepId==NOT_DRILLED_BY_EXPERT_STEP_ID)
					continue;

				// add new step id
				if(std::find(uniqueExpertDrillingStepsIds.begin(),uniqueExpertDrillingStepsIds.end(),expertDrillingStepId) == uniqueExpertDrillingStepsIds.end())
					uniqueExpertDrillingStepsIds.push_back(expertDrillingStepId);
			}
	// sort
	std::sort(uniqueExpertDrillingStepsIds.begin(), uniqueExpertDrillingStepsIds.end());

	if(uniqueExpertDrillingStepsIds.size())
		// the highest id found represents the total number of steps
		noOfDrillingSteps = uniqueExpertDrillingStepsIds[uniqueExpertDrillingStepsIds.size()-1];
	else
		noOfDrillingSteps = 1;


	drillingStepData_t drillingStepData;
	drillingStepData.startTime = 0.0;
	drillingStepData.finishTime = 0.0;


	for(unsigned char i = 1; i<=noOfDrillingSteps;i++)
	{
		for(unsigned int j=0; j<materialSegmentationModelNode->getNoOfSegments();j++)
		{
			cout <<"Adding data"<<endl;
			drillingStepData.noOfExpertVoxelsDrilledByUser.push_back(0);
			drillingStepData.noOfNonExpertVoxelsDrilledByUser.push_back(0);
			drillingStepData.noOfVoxelsDrilledByExpert.push_back(0);
		}

		this->drillingStepData.push_back(drillingStepData);
	}


	// loop through all voxels AGAIN!!! necessary to count the no of expert voxels in each segment
	for(unsigned int i =0;i<volumeModelNode->getWidth();i++)
		for(unsigned int j =0;j<volumeModelNode->getHeight();j++)
			for(unsigned int k =0;k<volumeModelNode->getDepth();k++)
			{
				Vec3i vec(i,j,k);
				unsigned char expertDrillingStepId = volumeModelNode->getExpertDrillingStepId(vec);

				// ignore non drilled voxels
				if(expertDrillingStepId==NOT_DRILLED_BY_EXPERT_STEP_ID)
					continue;

				incrementNoOfVoxelsDrilledByExpert(vec,expertDrillingStepId);

			}


	cout << "Number of Drilling Steps: "<< (unsigned int)noOfDrillingSteps<<endl;

	currentDrillingStepId = 1;
	cout << "Current Step Id: "<< (unsigned int)currentDrillingStepId<<endl;

	engageHighlightField->setValue(false,this->id);


	cout << "Not in Edit Expert Drilling Step Mode"<<endl;
	editExpertDrillingStepsMode = false;

	hasUserBoredThisDrillingStep = false;

	X3DChildNode::initialize();

	cout<<"Initialized Drilling Steps Node"<<endl<<endl;
}

bool DrillingStepsNode::isCurrentExpertDrillingStep(Vec3i p)
{
	if(volumeModelNode->getExpertDrillingStepId(p)==currentDrillingStepId)
		return true;
	return false;
}

bool DrillingStepsNode::isExpertDrillingStep(Vec3i p, unsigned char expertDrillingStepId)
{
	return volumeModelNode->getExpertDrillingStepId(p)==expertDrillingStepId?true:false;
}

const vector<Vec3i> DrillingStepsNode::getVoxelsDrilledByUserForCurrentDrillingStep()
{
	return getVoxelsDrilledByUserForDrillingStep(currentDrillingStepId);
}

const vector<Vec3i> DrillingStepsNode::getVoxelsDrilledByUserForDrillingStep(unsigned char drillingStepId)
{
	vector<Vec3i> voxelsDrilledByUserForDrillingStep;

	multimap<double, Vec3i>::iterator firstVoxel, lastVoxel, i;
	firstVoxel = volumeModelNode->timestamp_map.lower_bound( drillingStepData[drillingStepId-1].startTime);
	lastVoxel = volumeModelNode->timestamp_map.upper_bound(drillingStepData[drillingStepId-1].finishTime);

	for ( i=firstVoxel ; i != lastVoxel; i++ )
	{
		const Vec3i v = (*i).second;
		voxelsDrilledByUserForDrillingStep.push_back(v);
	}

	std::sort(voxelsDrilledByUserForDrillingStep.begin(),voxelsDrilledByUserForDrillingStep.end());

	return voxelsDrilledByUserForDrillingStep;
}

const vector<Vec3i> DrillingStepsNode::getVoxelsDrilledByExpertForCurrentDrillingStep()
{
	return getVoxelsDrilledByExpertForDrillingStep(currentDrillingStepId);
}

const vector<Vec3i> DrillingStepsNode::getVoxelsDrilledByExpertForDrillingStep(unsigned char drillingStepId)
{
	vector<Vec3i> voxelsDrilledByExpertForDrillingStep;

	for(unsigned int i =0;i<volumeModelNode->getWidth();i++)
		for(unsigned int j =0;j<volumeModelNode->getHeight();j++)
			for(unsigned int k =0;k<volumeModelNode->getDepth();k++)
			{
				Vec3i vec(i,j,k);

				if(isExpertDrillingStep(vec, drillingStepId))
				{
					voxelsDrilledByExpertForDrillingStep.push_back(vec);
				}
			}

	std::sort(voxelsDrilledByExpertForDrillingStep.begin(),voxelsDrilledByExpertForDrillingStep.end());

	return voxelsDrilledByExpertForDrillingStep;
}

void DrillingStepsNode::setAsBored(Vec3i v, double elapsedTime)
{
	// If we are in edit more then we will set this voxel as belonging to the current expert drilling step
	if(editExpertDrillingStepsMode)
	{
		volumeModelNode->setExpertDrillingStepId(v,currentDrillingStepId);
		incrementNoOfVoxelsDrilledByExpert(v, currentDrillingStepId);
	}

	if(!hasUserBoredThisDrillingStep)
		drillingStepData[currentDrillingStepId-1].startTime = elapsedTime;

	hasUserBoredThisDrillingStep = true;

	drillingStepData[currentDrillingStepId-1].finishTime = elapsedTime;

	//Have we have drilled in a voxel which belongs to the current drilling step
	if(isCurrentExpertDrillingStep(v))
	{
		incrementNoOfExpertVoxelsDrilledByUser(v, currentDrillingStepId);
	}
	else
	{
		incrementNoOfNonExpertVoxelsDrilledByUser(v, currentDrillingStepId);
	}
}

void DrillingStepsNode::setExpertDrillingStepId(Vec3i p, unsigned char step)
{
	if(step>noOfDrillingSteps)
	{
		stringstream failStringStream;
		failStringStream << "The Expert Drilling Step Id: " << (unsigned int)step << " is invalid";
		throw DrillingStepsNodeException(failStringStream.str());
	}

	volumeModelNode->setExpertDrillingStepId(p,step);
}

void DrillingStepsNode::saveExpertDrillingSteps()
{
	if(editExpertDrillingStepsMode)
	{
		if(volumeModelNode->expertDrillingStepsField->IsInitializedFromNrrdFile())
		{
			cout<<"Saving Expert Drilling Steps"<< endl;
			volumeModelNode->expertDrillingStepsField->saveDataToNrrdFile();
		}
		else
			cout<<"Drilled By has no associated file therefore expert cannot be saved"<< endl;
	}
	else
		cout<<"Not in expert drilling mode"<< endl;
}

void DrillingStepsNode::incrementCurrentDrillingStep()
{
	hasUserBoredThisDrillingStep = false;
	currentDrillingStepId = (currentDrillingStepId%noOfDrillingSteps)+1;

	cout << "Current Drilling Step Id: "<< (unsigned int)currentDrillingStepId<<endl;
	cout << "StartTime: " << drillingStepData[currentDrillingStepId-1].startTime<<endl;
	cout << "FinishTime: " << drillingStepData[currentDrillingStepId-1].finishTime<<endl;

	// Inform the highlight module to redraw information based on new expert step
	engageHighlightField->setValue(true,this->id);
	forbiddenSegmentationModelNode->incrementCurrentDrillingStep();
}

void DrillingStepsNode::decrementCurrentDrillingStep()
{
	hasUserBoredThisDrillingStep = false;

	if(currentDrillingStepId==1)
		currentDrillingStepId = noOfDrillingSteps;
	else
		currentDrillingStepId -= 1;

	cout << "Current Drilling Step Id: "<< (unsigned int)currentDrillingStepId<<endl;
	cout << "StartTime: " << drillingStepData[currentDrillingStepId-1].startTime<<endl;
	cout << "FinishTime: " << drillingStepData[currentDrillingStepId-1].finishTime<<endl;

	// Inform the highlight module to redraw information based on new expert step
	engageHighlightField->setValue(true,this->id);
	forbiddenSegmentationModelNode->decrementCurrentDrillingStep();
}

void DrillingStepsNode::addDrillingStep()
{
	if(editExpertDrillingStepsMode)
	{
		if(noOfDrillingSteps!=UCHAR_MAX)
		{
			noOfDrillingSteps++;

			materialSegmentationModelNode = static_cast< MaterialSegmentationModel * >( materialSegmentationModelField->getValue() );

			drillingStepData_t drillingStepData;

			for(int j=0; j<(int)materialSegmentationModelNode->getNoOfSegments();j++)
			{
				drillingStepData.noOfExpertVoxelsDrilledByUser.push_back(0);
				drillingStepData.noOfNonExpertVoxelsDrilledByUser.push_back(0);
				drillingStepData.noOfVoxelsDrilledByExpert.push_back(0);
			}

			this->drillingStepData.push_back(drillingStepData);

			cout << "Adding Expert Drilling Step: "<< (unsigned int)noOfDrillingSteps<<endl;

			currentDrillingStepId = noOfDrillingSteps;

			cout << "Current Expert Drilling Step Id: "<< (unsigned int)currentDrillingStepId<<endl;

			// Inform the highlight module to redraw information based on new expert step
			engageHighlightField->setValue(true,this->id);
		}
		else
			cout << "Unable to add a new Expert Drilling Step since the maximum has been reached";
	}
	else
		cout<<"Not in expert drilling mode"<< endl;
}

void DrillingStepsNode::clearCurrentDrillingStep()
{
	if(editExpertDrillingStepsMode)
	{
		// loop through all voxels
		for(unsigned int i=0;i<volumeModelNode->getWidth();i++)
			for(unsigned int j=0;j<volumeModelNode->getHeight();j++)
				for(unsigned int k=0;k<volumeModelNode->getDepth();k++)
				{
					Vec3i vec(i,j,k);
					if(isCurrentExpertDrillingStep(vec))
						volumeModelNode->setExpertDrillingStepId(vec, NOT_DRILLED_BY_EXPERT_STEP_ID);
				}

		materialSegmentationModelNode = static_cast< MaterialSegmentationModel * >( materialSegmentationModelField->getValue() );

		for(int j=0; j<(int)materialSegmentationModelNode->getNoOfSegments();j++)
		{
			drillingStepData[currentDrillingStepId-1].noOfExpertVoxelsDrilledByUser.setValue(j,0);
			drillingStepData[currentDrillingStepId-1].noOfNonExpertVoxelsDrilledByUser.setValue(j,0);
			drillingStepData[currentDrillingStepId-1].noOfVoxelsDrilledByExpert.setValue(j,0);
		}


		cout << "Clearing Expert Drilling Step with Id: "<< (unsigned int)currentDrillingStepId<<endl;
		// Inform the highlight module to redraw information based on new expert step
		engageHighlightField->setValue(true,this->id);

	}
	else
		cout<<"Not in expert drilling mode"<< endl;
}

void DrillingStepsNode::enterEditExpertDrillingStepsMode(bool enterEditExpertDrillingStepsMode)
{
	editExpertDrillingStepsMode = enterEditExpertDrillingStepsMode;
}



float DrillingStepsNode::fractionOfVoxelsDrilledByUserToExpertForCurrentDrillingStep(){

	int noOfExpertVoxelsDrilledByUser=0;
	int noOfVoxelsDrilledByExpert=0;

	materialSegmentationModelNode = static_cast< MaterialSegmentationModel * >( materialSegmentationModelField->getValue() );

	for(int j=0; j<(int)materialSegmentationModelNode->getNoOfSegments();j++)
	{
		noOfExpertVoxelsDrilledByUser += drillingStepData[currentDrillingStepId-1].noOfExpertVoxelsDrilledByUser.getValue()[j];
		noOfVoxelsDrilledByExpert +=     drillingStepData[currentDrillingStepId-1].noOfVoxelsDrilledByExpert.getValue()[j];
	}



	float fraction = 1;

	if(noOfVoxelsDrilledByExpert!=0)
		fraction = (float)noOfExpertVoxelsDrilledByUser/(float)noOfVoxelsDrilledByExpert;

	return fraction;
}


int DrillingStepsNode::noOfNonExpertVoxelsDrilledByUserForCurrentDrillingStep(int segmentID){
	return drillingStepData[currentDrillingStepId-1].noOfNonExpertVoxelsDrilledByUser.getValue()[segmentID];
}

int DrillingStepsNode::noOfExpertVoxelsDrilledByUserForCurrentDrillingStep(int segmentID){
	return drillingStepData[currentDrillingStepId-1].noOfExpertVoxelsDrilledByUser.getValue()[segmentID];
}

int DrillingStepsNode::noOfVoxelsDrilledByExpertForCurrentDrillingStep(int segmentID){
	return drillingStepData[currentDrillingStepId-1].noOfVoxelsDrilledByExpert.getValue()[segmentID];
}

int DrillingStepsNode::noOfNonExpertVoxelsDrilledByUser(int segmentID, int drillingStepId){
	return drillingStepData[drillingStepId-1].noOfNonExpertVoxelsDrilledByUser.getValue()[segmentID];
}

int DrillingStepsNode::noOfExpertVoxelsDrilledByUser(int segmentID, int drillingStepId){
	return drillingStepData[drillingStepId-1].noOfExpertVoxelsDrilledByUser.getValue()[segmentID];
}

int DrillingStepsNode::noOfVoxelsDrilledByExpert(int segmentID, int drillingStepId){
	return drillingStepData[drillingStepId-1].noOfVoxelsDrilledByExpert.getValue()[segmentID];
}



void DrillingStepsNode::incrementNoOfVoxelsDrilledByExpert(Vec3i v, int drillingStepId){
	int noOfVoxelsDrilled = this->drillingStepData[drillingStepId-1].
	noOfVoxelsDrilledByExpert.getValue()[volumeModelNode->getMaterialSegmentationId(v)];

	drillingStepData[drillingStepId-1].noOfVoxelsDrilledByExpert.setValue(
			volumeModelNode->getMaterialSegmentationId(v),++noOfVoxelsDrilled);

}

void DrillingStepsNode::incrementNoOfExpertVoxelsDrilledByUser(Vec3i v, int drillingStepId){
	int noOfVoxelsDrilled = this->drillingStepData[drillingStepId-1].
	noOfExpertVoxelsDrilledByUser.getValue()[volumeModelNode->getMaterialSegmentationId(v)];

	drillingStepData[drillingStepId-1].noOfExpertVoxelsDrilledByUser.setValue(
			volumeModelNode->getMaterialSegmentationId(v),++noOfVoxelsDrilled);
}

void DrillingStepsNode::incrementNoOfNonExpertVoxelsDrilledByUser(Vec3i v, int drillingStepId){
	int noOfVoxelsDrilled = this->drillingStepData[drillingStepId-1].
	noOfNonExpertVoxelsDrilledByUser.getValue()[volumeModelNode->getMaterialSegmentationId(v)];

	drillingStepData[drillingStepId-1].noOfNonExpertVoxelsDrilledByUser.setValue(
			volumeModelNode->getMaterialSegmentationId(v),++noOfVoxelsDrilled);
}


unsigned char DrillingStepsNode::getNoOfDrillingSteps(){
	return noOfDrillingSteps;
}

unsigned char DrillingStepsNode::getCurrentDrillingStepId()
{
	return currentDrillingStepId;
}
