/**
 * DrillingStepsNode.h
 *
 *  Created on: Jan 27, 2009
 *      Author: barrygr
 */

#ifndef DILLING_STEPS_NODE_H_
#define DILLING_STEPS_NODE_H_

#include <H3D/X3DChildNode.h>
#include <H3D/H3DImageObject.h>
#include <H3D/MFInt32.h>
#include <H3D/MFFloat.h>
#include <H3D/SFVec3f.h>
#include <H3D/X3DTexture3DNode.h>
#include <H3D/Image3DTexture.h>
#include <H3DUtil/Image.h>
#include <H3DUtil/Vec3f.h>
#include <H3DUtil/TimeStamp.h>
#include <HAPI/HAPIForceEffect.h>
#include "MFUnsignedInt8.h"
#include "MFUnsignedInt8FromNrrdFile.h"
#include "MFFloatFromNrrdFile.h"
#include "MFDoubleFromNrrdFile.h"
#include "MFieldFromNrrdFile.h"
#include "VolumeModel.h"
#include "PartitionModel.h"
#include "MaterialSegmentationModel.h"
#include "ForbiddenSegmentationModel.h"
#include "Vec3i.h"
#include <map>
#include <assert.h>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <ctime>

using namespace std;
using namespace H3D;

namespace FS
{

#define NOT_DRILLED_BY_EXPERT_STEP_ID 			0

class VolumeModel;

/**
 * A drilling step stores a set of voxels as drilled by an expert dentist. Each step represents a particular step
 * within an dental operation. For example, one step might be the removal of the tooth's crown
 */
class DrillingStepsNode : public H3D::X3DChildNode
{
public:

	/**
	 * Information for each drilling step
	 */
	struct drillingStepData_t
	{
		/// The time that the student began drilling for this drilling step
		double startTime;
		/// The time that the student finished drilling for this drilling step
		double finishTime;
		/// The number of voxels removed by the expert for this drilling step and segment
		H3D::MFInt32 noOfVoxelsDrilledByExpert;
		/// The number of voxels that the user has managed to drill correctly for this drilling step and segment
		H3D::MFInt32  noOfExpertVoxelsDrilledByUser;
		/// The number of voxels that the user has managed to drill correctly for this drilling step and segment
		H3D::MFInt32  noOfNonExpertVoxelsDrilledByUser;
	};

	/**
	 *\param The fraction of voxels drilled by the the student with respect to the expert for the current drilling step, OUTPUT_ONLY
	 *\param Informs Highlight Node that it is necessary to redraw the image, OUTPUT_ONLY
	 *\param Requires access to VolumeModelNode, INITIALIZE_ONLY
	 */
	DrillingStepsNode(
			H3D::Inst< H3D::SFBool >    engageHighlightField = 0,
			H3D::Inst< H3D::SFNode >    volumeModelField = 0,
			H3D::Inst< H3D::SFNode >    materialSegmentationModelField = 0,
   			H3D::Inst< H3D::SFNode >    forbiddenSegmentationModelField = 0
			);

	virtual ~DrillingStepsNode()
	{
		cout<<"Drilling Steps Node Destroyed"<<endl;
	}

	void initialize();

	/**
	 * The H3DNodeDatabase for this node.
	 */
	static H3D::H3DNodeDatabase database;

	/**
	 * We must inform the Highlight Node that it needs to redraw the screen image for example. OUTPUT_ONLY Field
	 * when we step between steps
	 */
	auto_ptr< H3D::SFBool >                     engageHighlightField;

	/**
	 * Access to Volume Model. INITIALIZE_ONLY Field
	 */
	auto_ptr< H3D::SFNode > 					volumeModelField;

	/**
	 * Access to material segmentation Model. INITIALIZE_ONLY Field
	 */
	auto_ptr< H3D::SFNode > 					materialSegmentationModelField;

	/**
	 * Access to forbidden segmentation Model. INITIALIZE_ONLY Field
	 */
	auto_ptr< H3D::SFNode > 					forbiddenSegmentationModelField;


	/**
	 * Return the nr of voxels drilled by the User which is not Expert voxels for the Currently selected Drilling Step
	 * \return No of voxels drilled User and not by the Expert
	 */
	int noOfNonExpertVoxelsDrilledByUserForCurrentDrillingStep(int segmentID);

	/**
	 * Return the nr of voxels drilled by the User vs Expert for the Currently selected Drilling Step
	 * \return Fraction of voxels drilled User vs Expert
	 */
	int noOfExpertVoxelsDrilledByUserForCurrentDrillingStep(int segmentID);

	/**
	 * Return the nr of voxels drilled by Expert for the Currently selected Drilling Step
	 * \return Fraction of voxels drilled User vs Expert
	 */
	int noOfVoxelsDrilledByExpertForCurrentDrillingStep(int segmentID);


	/**
	 * Return the nr of voxels drilled by the User but not by the Expert drilling Step and segment
	 * \return Nr of voxels drilled User but not by Expert
	 */
	int noOfNonExpertVoxelsDrilledByUser(int segmentID, int drillingStepId);

	/**
	 * Return the nr of voxels drilled by the User vs Expert drilling Step and segment
	 * \return Nr of voxels drilled User vs Expert
	 */
	int noOfExpertVoxelsDrilledByUser(int segmentID, int drillingStepId);

	/**
	 * Return the nr of voxels drilled by Expert for the selected Drilling Step and segment
	 * \return Nr of voxels drilled User vs Expert
	 */
	int noOfVoxelsDrilledByExpert(int segmentID, int drillingStepId);



	/**
	 * Return the fraction of voxels drilled by the User vs Expert for the Currently selected Drilling Step
	 * \return Fraction of voxels drilled User vs Expert
	 */
	float fractionOfVoxelsDrilledByUserToExpertForCurrentDrillingStep();

	/**
	 * Return the voxels drilled by the User for the Currently selected Drilling Step
	 * \return The sorted list of voxels
	 */
	const vector<Vec3i> getVoxelsDrilledByUserForCurrentDrillingStep();

	/**
	 * Return the voxels drilled by the Expert for the Currently selected Drilling Step
	 * \return The sorted list of voxels
	 */
	const vector<Vec3i> getVoxelsDrilledByExpertForCurrentDrillingStep();

	/**
	 * Increment the array of expert drilled voxels for the specific segment linked to the voxel at v for the drillingstep
	 */
	void incrementNoOfVoxelsDrilledByExpert(Vec3i v, int drillStepId);

	/**
	 * Increment the array of user drilled expert voxels for the specific segment linked to the voxel at v for the drillingstep
	 */
	void incrementNoOfExpertVoxelsDrilledByUser(Vec3i v, int drillStepId);

	/**
	 * Increment the array of user drilled expert voxels for the specific segment linked to the voxel at v for the drillingstep
	 */
	void incrementNoOfNonExpertVoxelsDrilledByUser(Vec3i v, int drillStepId);

	/**
	 * Set that we have removed this voxel. If we are in Expert Edit Mode then this voxel will be added to the list of voxels
	 * drilled by the expert for this Drilling Step. Either way, this voxel will be added to the list of voxels drilled by
	 * the student
	 */
	void setAsBored(Vec3i v, double elapsedTime);

	/**
	 * Save expert drilled steps to nrrd file. Must be in Expert Drilling Step Mode
	 */
	void saveExpertDrillingSteps();

	/**
	 * Move to the next Drilling Step in the sequence. Loops back to the start once the end is reached
	 */
	void incrementCurrentDrillingStep();

	/**
	 * Move to the previous Drilling Step in the sequence. Jumps to the end of the sequence once the start is reached
	 */
	void decrementCurrentDrillingStep();

	/**
	 * And a drilling step to the end of the Drilling Stpes sequence.  Must be in Expert Drilling Step Mode
	 */
	void addDrillingStep();

	/**
	 * Clear the user and expert data stored for the current drilling step. Must be in Expert Drilling Step Mode
	 */
	void clearCurrentDrillingStep();

	/**
	 * User to enter or exit Expert Drilling Steps Mode
	 *\param true if we wish to enter Expert Drilling Steps Mode
	 */
	void enterEditExpertDrillingStepsMode(bool enterEditExpertDrillingStepsMode);

	/**
	 * Number of drillingsteps in current model
	 */
	unsigned char getNoOfDrillingSteps();

	unsigned char getCurrentDrillingStepId();

	const vector<Vec3i> getVoxelsDrilledByExpertForDrillingStep(unsigned char drillingStepId);

	const vector<Vec3i> getVoxelsDrilledByUserForDrillingStep(unsigned char drillingStepId);

	/**
	 * Exception class for DrillingStepsNode
	 */
	class DrillingStepsNodeException: public runtime_error
	{
	public:

		/**
		 * Constructor
		 */
		DrillingStepsNodeException(const string& what) :
			runtime_error("Exception within DrillingStepsNode: " + what)
			{
			}
	};

private:
	VolumeModel *volumeModelNode;
	MaterialSegmentationModel *materialSegmentationModelNode;
	ForbiddenSegmentationModel *forbiddenSegmentationModelNode;
	unsigned char noOfDrillingSteps; 			// The number of drilling steps
	std::vector<drillingStepData_t> drillingStepData;
	unsigned char currentDrillingStepId;		// The currently selected drilling step
	bool hasUserBoredThisDrillingStep;
	bool isExpertDrillingStep(Vec3i p, unsigned char expertDrillingStepId);
	bool isCurrentExpertDrillingStep(Vec3i p);
	void setExpertDrillingStepId(Vec3i p, unsigned char step);
	bool editExpertDrillingStepsMode;			// true if we wish to enter Expert Drilling Steps Mode
};
}

#endif /* DILLING_STEPS_NODE_H_ */
