/**
 * ForbiddenSegmentationModel.cpp
 *
 *  Created on: Mar 23, 2009
 *      Author: barrygr
 */

#include "ForbiddenSegmentationModel.h"

using namespace FS;

// Add this node to the H3DNodeDatabase system.
H3D::H3DNodeDatabase ForbiddenSegmentationModel::database(
		"ForbiddenSegmentationModel",
		&(newInstance<ForbiddenSegmentationModel>),
		typeid(ForbiddenSegmentationModel),
		&X3DChildNode::database);

namespace ForbiddenSegmentationModelInternals
{
H3D::FIELDDB_ELEMENT(ForbiddenSegmentationModel, segmentIdField, INITIALIZE_ONLY );
H3D::FIELDDB_ELEMENT(ForbiddenSegmentationModel, segmentNameField, INITIALIZE_ONLY );
H3D::FIELDDB_ELEMENT(ForbiddenSegmentationModel, noOfVoxelsBoredByUser, OUTPUT_ONLY);
}

ForbiddenSegmentationModel::ForbiddenSegmentationModel(
		H3D::Inst< MFUnsignedInt8 >		 		segmentIdField,
		H3D::Inst< H3D::MFString > 				segmentNameField,
		H3D::Inst< H3D::MFInt32 > 				noOfVoxelsBoredByUser)
:segmentIdField(segmentIdField),
segmentNameField(segmentNameField),
noOfVoxelsBoredByUser(noOfVoxelsBoredByUser)
{
	cout << "Forbidden Segmentation Model Node Constructing"<<endl;

	type_name = "ForbiddenSegmentationModel";
	database.initFields( this );

	cout << "Forbidden Segmentation Model Node Constructed"<<endl<<endl;
}

void ForbiddenSegmentationModel::initialize()
{
	cout<<"Forbidden Segmentation Model Initializing"<<endl;

	// check if the number of segments defined within the segmentIdField is greater than zero
	if(!(segmentIdField->getValue()).size())
		throw ForbiddenSegmentationModelException("partitionIdField should contain one or more values");

	// UNDEFINED_FORBIDDEN_SEGMENT_ID is reserved
	if(existsInSegmentIdField(UNDEFINED_FORBIDDEN_SEGMENT_ID))
	{
		stringstream failStringStream;
		failStringStream << "The Id: " << (int)UNDEFINED_FORBIDDEN_SEGMENT_ID <<" is reserved for " << UNDEFINED_FORBIDDEN_SEGMENT_NAME;
		throw ForbiddenSegmentationModelException(failStringStream.str());
	}

	if(segmentIdField->getSize() != segmentNameField->getSize())
		throw ForbiddenSegmentationModelException("segmentNameField should contain the same number of values as segmentIdField");

	// UNDEFINED_FORBIDDEN_SEGMENT_NAME is reserved
	if(existsInSegmentNameField(UNDEFINED_FORBIDDEN_SEGMENT_NAME))
	{
		stringstream failStringStream;
		failStringStream << "The Name: " << UNDEFINED_FORBIDDEN_SEGMENT_NAME <<" is reserved";
		throw ForbiddenSegmentationModelException(failStringStream.str());
	}

	// initialize the noOfVoxelsBoredByUser list with zeros
	//noOfVoxelsBoredByUser is kept for now. It is used (only?) for displaying number
	//of voxels removed in the GUI (the GUI should use vForbiddenDrillingStepData).
	//This is total number removed, not removed in each drillstep.
	unsigned int size = (segmentIdField->getValue()).size()+1;
	H3DInt32* data2 = new H3DInt32[size];
	for(unsigned int i = 0; i< size;i++)
		data2[i] = 0;
	noOfVoxelsBoredByUser->setValueFromVoidPtr(data2,size,size*sizeof(H3DInt32),this->id);
	delete[] data2;

	//A vector with struct containing drilled data in each drill step.
	forbiddenDrillingStepData_t sForbiddenDrillingStepData;
	noOfDrillingSteps = 2;

	for(unsigned char i = 1; i<=noOfDrillingSteps;i++)
	{
		cout <<"Adding initializingdata to struct, Forbidden Segmentation Model "<<endl;
		H3DInt32* data = new H3DInt32[size];
		for(unsigned int i = 0; i< size;i++)
			data[i] = 0;
		sForbiddenDrillingStepData.noOfForbiddenVoxelsBoredByUser.setValueFromVoidPtr(data,size,size*sizeof(H3DInt32),this->id);
		delete[] data;
		this->vForbiddenDrillingStepData.push_back(sForbiddenDrillingStepData);
	}

	currentDrillingStepId = 1;

	X3DChildNode::initialize();
	cout<<"Forbidden Segmentation Model Initialized"<<endl;
}

ForbiddenSegmentationModel::~ForbiddenSegmentationModel()
{
	cout<<"Forbidden Segmentation Model Detroyed"<<endl;
}

void ForbiddenSegmentationModel::incrementNoOfVoxelsBoredByUser(const unsigned char& segmentId)
{
	unsigned char index;

	// were not interested in the undefined voxels
	if(segmentId==UNDEFINED_FORBIDDEN_SEGMENT_ID)
		index = 0;
	else
		index = getSegmentIdFieldIndex(segmentId)+1;

	unsigned int currentValue2 = noOfVoxelsBoredByUser->getValue(this->id)[index];
	noOfVoxelsBoredByUser->setValue(index, ++currentValue2,this->id);

	unsigned int currentValue = this->vForbiddenDrillingStepData[currentDrillingStepId-1].noOfForbiddenVoxelsBoredByUser.getValue(this->id)[index];
	this->vForbiddenDrillingStepData[currentDrillingStepId-1].noOfForbiddenVoxelsBoredByUser.setValue(index, ++currentValue,this->id);
}

const std::vector<unsigned char> ForbiddenSegmentationModel::getSegmentIds()
{
	vector<unsigned char> segmentIds;

	segmentIds.push_back(UNDEFINED_FORBIDDEN_SEGMENT_ID);
	vector<unsigned char>::const_iterator i;
	for(i=segmentIdField->getValue().begin();i<segmentIdField->getValue().end();i++)
		segmentIds.push_back(*i);

	return segmentIds;
}

const bool ForbiddenSegmentationModel::existsInSegmentIdField(const unsigned char& segmentId)
{
	MFUnsignedInt8::const_iterator segmentIdFieldIterator;

	for(segmentIdFieldIterator=segmentIdField->begin();
		segmentIdFieldIterator<segmentIdField->end();
		segmentIdFieldIterator++)
	{
		if(*segmentIdFieldIterator==segmentId)
			return true;
	}

	return false;
}

const unsigned char ForbiddenSegmentationModel::getSegmentIdFieldIndex(const unsigned char& segmentId)
{
	MFUnsignedInt8::const_iterator segmentIdFieldIterator;
	unsigned char index;

	for(segmentIdFieldIterator=segmentIdField->begin(), index = 0;
		segmentIdFieldIterator<segmentIdField->end();
		segmentIdFieldIterator++, index++)
	{
		if(*segmentIdFieldIterator==segmentId)
			return index;
	}

	stringstream failStringStream;
	failStringStream << "Can't get the index for the Segment with Id: " << (int)segmentId <<" since it is not defined in segmentIdField";
	throw ForbiddenSegmentationModelException(failStringStream.str());
}


const bool ForbiddenSegmentationModel::existsInSegmentNameField(const string& segmentName)
{
	MFString::const_iterator segmentNameFieldIterator;

	for(segmentNameFieldIterator=segmentNameField->begin();
		segmentNameFieldIterator<segmentNameField->end();
		segmentNameFieldIterator++)
	{
		if(*segmentNameFieldIterator==segmentName)
			return true;
	}

	return false;
}

void ForbiddenSegmentationModel::incrementCurrentDrillingStep()
{
	//currentDrillingStepId = (currentDrillingStepId%noOfDrillingSteps)+1;
	currentDrillingStepId++;
}

void ForbiddenSegmentationModel::decrementCurrentDrillingStep()
{
	//currentDrillingStepId = (currentDrillingStepId%noOfDrillingSteps)+1;
	currentDrillingStepId--;
}

int ForbiddenSegmentationModel::noOfForbiddenVoxelsDrilledByUser(int segmentID, int drillingStepId)
{
	return vForbiddenDrillingStepData[drillingStepId-1].noOfForbiddenVoxelsBoredByUser.getValue()[segmentID];
}

