/**
 * ForbiddenSegmentationModel.h
 *
 *  Created on: Mar 23, 2009
 *      Author: barrygr
 */

#ifndef FORBIDDENSEGMENTATIONMODEL_H_
#define FORBIDDENSEGMENTATIONMODEL_H_

#include <H3D/X3DChildNode.h>
#include <H3D/MFString.h>
#include <H3D/MFInt32.h>
#include "MFUnsignedInt8.h"

using namespace std;
using namespace H3D;

namespace FS
{

#define UNDEFINED_FORBIDDEN_SEGMENT_ID 0	// Voxels not belonging to any Forbidden Segment
#define UNDEFINED_FORBIDDEN_SEGMENT_NAME "Undefined Forbidden Segment"

/**
 * \brief Defines a groups of segments in which the user is not supposed to bore
 */
class ForbiddenSegmentationModel : public H3D::X3DChildNode
{
public:

	/**
	 * Information about Forbidden segments drilled in each drilling step
	 */
	struct forbiddenDrillingStepData_t
	{
		//The amount bored in each Forbidden segment by the user.
		H3D::MFInt32 noOfForbiddenVoxelsBoredByUser;
	};

	/**
	 * Constructor
	 * \param segmentIdField A list of Forbidden Segment Ids, INITIALIZE_ONLY Field
	 * \param segmentNameField A list of Forbidden Segment Names, INITIALIZE_ONLY Field
	 * \param noOfVoxelsBoredByUser The amount bored from each segment by the user, OUTPUT_ONLY Field
	 */
	ForbiddenSegmentationModel(
			H3D::Inst< MFUnsignedInt8 >		 		segmentIdField = 0,
			H3D::Inst< H3D::MFString > 				segmentNameField = 0,
			H3D::Inst< H3D::MFInt32 > 				noOfVoxelsBoredByUser = 0
	);

	virtual ~ForbiddenSegmentationModel();

	void initialize();

	/**
	 * Increment by one the number of voxels drilled by the user for a particular segment
	 * \param segmentId The segment in question
	 */
	void incrementNoOfVoxelsBoredByUser(const unsigned char& segmentId);

	/**
	 * Return the list of Segment Ids associated with the Forbidden Segment Model. The UNDEFINED_FORBIDDEN_SEGMENT segment is the first segment in this list
	 * \return The Segment Ids
	 */
	const std::vector<unsigned char> getSegmentIds();

	/**
	 * Increment by one the current drilling step.
	 */
	void incrementCurrentDrillingStep();

	/**
	 * Decrement by one the current drilling step.
	 */
	void decrementCurrentDrillingStep();

	/**
		 * Returns the number of voxels drilled by the user in each drillstep for each forbidden segment,
		 * with the undefined segment value stored as the first value in the noOfForbiddenVoxelsBoredByUser field.
		 */
	int noOfForbiddenVoxelsDrilledByUser(int segmentID, int drillingStepId);

	/**
	 * The H3DNodeDatabase for this node.
	 */
	static H3D::H3DNodeDatabase database;

	/**
	 * The list of Forbidden Segment Ids
	 */
	auto_ptr< MFUnsignedInt8 >	segmentIdField;

	/**
	 * The list of Forbidden Segment Names
	 */
	auto_ptr< H3D::MFString >		segmentNameField;

	/**
	 * The number of voxels drilled by the user for each forbidden segment with the undefined segment value
	 * stored as the first value in the noOfVoxelsBoredByUser field
	 */
	auto_ptr< H3D::MFInt32 >		noOfVoxelsBoredByUser;

	/**
	 * Exception class for Forbidden Segmentation Model
	 */
	class ForbiddenSegmentationModelException: public runtime_error
	{
	public:

		/**
		 * Constructor
		 */
		ForbiddenSegmentationModelException(const string& what) :
			runtime_error("Exception within ForbiddenSegmentationModel: " + what)
			{
			}
	};

private:
	const bool existsInSegmentIdField(const unsigned char& segmentId);
	const bool existsInSegmentNameField(const string& segmentName);
	const unsigned char getSegmentIdFieldIndex(const unsigned char& segmentId);
	std::vector<forbiddenDrillingStepData_t> vForbiddenDrillingStepData; //Vector with amount bored in each Forbidden segment and in each drilling step by the user.
	int noOfDrillingSteps;
	int currentDrillingStepId;
};
}

#endif /* FORBIDDENSEGMENTATIONMODEL_H_ */
