//////////////////////////////////////////////////////////////////////////////
//    Copyright 2008 Forsslund Systems AB
//
//    ADrillableNode is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    ADrillableNode is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with ADrillableNode; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//    If you wish to use the code in a proprietary project, please contact
//    jonas@forsslundsystems.se for more information.
//
//////////////////////////////////////////////////////////////////////////////
#include "HighlightNode.h"

using namespace FS;

/// Add this node to the H3DNodeDatabase system.
H3D::H3DNodeDatabase HighlightNode::database( "HighlightNode",
		&(newInstance<HighlightNode>),
		typeid( HighlightNode ),
		&X3DChildNode::database );

namespace HighlightNodeInternals
{
H3D::FIELDDB_ELEMENT( HighlightNode, volumeModelField, INITIALIZE_ONLY );
H3D::FIELDDB_ELEMENT( HighlightNode, drillingStepsField, INITIALIZE_ONLY );
}

HighlightNode::HighlightNode(H3D::Inst< H3D::SFNode > volumeModelField,
		H3D::Inst< H3D::SFNode  > drillingStepsField):
			volumeModelField( volumeModelField ),
			drillingStepsField(drillingStepsField)
			{

	type_name = "HighlightNode";
	database.initFields( this );
			}

void HighlightNode::initialize()
{
	cout<<"Initializing Highlight Node"<<endl;

	if(volumeModelField->getValue()==NULL)
		throw HighlightNodeException("Expected the volumeModelField to be set within the x3d file");

	if(drillingStepsField->getValue()==NULL)
		throw HighlightNodeException("Expected the drillingStepsField to be set within the x3d file");

	volumeModelNode = static_cast< VolumeModel * >( volumeModelField->getValue() );
	drillingStepsNode = static_cast< DrillingStepsNode * >( drillingStepsField->getValue() );

	currentDrillingStep = 1;

	highlightExpertOn = false;

	X3DChildNode::initialize();

	cout<<"Initialized Highlight Node"<<endl;
}

void HighlightNode::highlightUser(bool enable)
{

	vector<Vec3i>::const_iterator i;
	const vector<Vec3i> voxelsDrilledByUserForCurrentStep = drillingStepsNode->getVoxelsDrilledByUserForCurrentDrillingStep();

	if(enable)
	{
		cout << "Enable Highlight User" << endl;
		for(i=voxelsDrilledByUserForCurrentStep.begin();i<voxelsDrilledByUserForCurrentStep.end();i++)
		{
			volumeModelNode->setImageValue(*i,1/255.0,0.0,0.0);
		}
	}

	else{
		cout << "Disable Highlight User" << endl;
		for(i=voxelsDrilledByUserForCurrentStep.begin();i<voxelsDrilledByUserForCurrentStep.end();i++)
		{
			if(volumeModelNode->getMaterialRemaining(*i)==0) // air at this point in time
				volumeModelNode->setImageValue(*i,0.0,0,0);
			else
				volumeModelNode->setImageValue(*i,volumeModelNode->getOriginalImageValue(*i),0,0);

		}
	}

}

void HighlightNode::highlightExpert(bool enable)
{

	vector<Vec3i>::const_iterator i;
	const vector<Vec3i> voxelsDrilledByExpertForCurrentStep = drillingStepsNode->getVoxelsDrilledByExpertForCurrentDrillingStep();

	if(enable)
	{
		cout << "Enable Highlight Expert" << endl;
		for(i=voxelsDrilledByExpertForCurrentStep.begin();i<voxelsDrilledByExpertForCurrentStep.end();i++)
		{
			if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
				volumeModelNode->setImageValue(*i,6/255.0,0.0,0.0);
		}
		highlightExpertOn = true;

	}
	else
	{
		cout << "Disable Highlight Expert" << endl;
		for(i=voxelsDrilledByExpertForCurrentStep.begin();i<voxelsDrilledByExpertForCurrentStep.end();i++)
		{
			if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
				volumeModelNode->setImageValue(*i,volumeModelNode->getOriginalImageValue(*i),0,0);
		}
		highlightExpertOn = false;
	}
}


void HighlightNode::highlightForbidden(bool enable)
{
	vector<Vec3i>::const_iterator i;
	const vector<Vec3i> forbiddenVoxels = volumeModelNode->getAllForbiddenVoxels();

	if(enable)
	{
		cout << "Enable Highlight Forbidden" << endl;
		for(i=forbiddenVoxels.begin();i<forbiddenVoxels.end();i++)
		{
			if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
				volumeModelNode->setImageValue(*i,3/255.0,0.0,0.0);
		}
	}
	else
	{
		cout << "Disable Highlight Forbidden" << endl;
		for(i=forbiddenVoxels.begin();i<forbiddenVoxels.end();i++)
		{
			if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
				volumeModelNode->setImageValue(*i,volumeModelNode->getOriginalImageValue(*i),0,0);
		}
	}
}

void HighlightNode::traverseSG( H3D::TraverseInfo &ti )
{
	if(currentDrillingStep != drillingStepsNode->getCurrentDrillingStepId())
	{
		if(highlightExpertOn)
		{
			vector<Vec3i>::const_iterator i;

			// Remove highlighted voxels for previous drillingstep
			const vector<Vec3i> voxelsDrilledByExpertForPreviousStep = drillingStepsNode->getVoxelsDrilledByExpertForDrillingStep(currentDrillingStep);
			for(i=voxelsDrilledByExpertForPreviousStep.begin();i<voxelsDrilledByExpertForPreviousStep.end();i++)
			{
				if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
						volumeModelNode->setImageValue(*i,volumeModelNode->getOriginalImageValue(*i),0,0);
			}

			// Highlight voxels in current drillingstep
			const vector<Vec3i> voxelsDrilledByExpertForCurrentStep = drillingStepsNode->getVoxelsDrilledByExpertForDrillingStep(currentDrillingStep + 1);
			for(i=voxelsDrilledByExpertForCurrentStep.begin();i<voxelsDrilledByExpertForCurrentStep.end();i++)
			{
				if(volumeModelNode->getMaterialRemaining(*i)!=0) // not air at this point in time
					volumeModelNode->setImageValue(*i,6/255.0,0.0,0.0);
			}
		}
		currentDrillingStep = drillingStepsNode->getCurrentDrillingStepId();
	}
}
