/**
 * StateMachine.cpp
 *
 *  Created on: March 25, 2009
 *      Author: parhe
 */

#ifndef STATEMACHINE_H_
#define STATEMACHINE_H_

#include <H3D/X3DChildNode.h>
#include <H3D/SFBool.h>
#include <H3D/SFFloat.h>
#include <H3D/SFInt32.h>
#include <H3D/MFFloat.h>
#include <H3D/MFString.h>
#include <H3D/SFVec3f.h>
#include <H3D/MFVec3f.h>
#include <H3D/SFRotation.h>
#include "SFUnsignedInt8.h"
#include "MFUnsignedInt8.h"
#include "VolumeModel.h"
#include "DrillingStepsNode.h"
#include <stdexcept>

namespace FS
{

/**
 * \brief The state machine of the simulator. Keeps track of which state user is in and control if
 * user is allowed to go to next state. Also removes crown and root when user is in correct state.
 * All prerequisites to move between states are set in X3D file!
 */
class StateMachine: public H3D::X3DChildNode
{

public:

	/**
	 * Constructor

	 */
	StateMachine(H3D::Inst< H3D::MFVec3f> _triggerPointsRequiredField = 0,
			H3D::Inst< H3D::MFFloat> _distancesRequiredField = 0,
			H3D::Inst< H3D::MFFloat> _forcesRequiredField = 0,
			H3D::Inst< H3D::MFInt32> _deviceShapesRequiredField = 0,
			H3D::Inst< H3D::MFFloat> _fractionOfExpertRequiredField = 0,
			H3D::Inst< H3D::MFFloat> _rotationAnglesRequiredField = 0,
			H3D::Inst< H3D::MFString> _partitionNamesField = 0,
			H3D::Inst< H3D::MFInt32> _drillingStepsField = 0,
			H3D::Inst< H3D::SFInt32> _currentDeviceShapeField = 0,
			H3D::Inst< H3D::SFVec3f> _currentForceField = 0,
			H3D::Inst< H3D::SFVec3f> _drillPosField = 0,
			H3D::Inst< H3D::SFVec3f> _currentRotationAnglesField = 0,
			H3D::Inst< H3D::SFInt32> _stateField = 0,
			H3D::Inst< H3D::SFNode > _volumeModelField = 0,
			H3D::Inst< H3D::SFNode > _drillingStepsNodeField = 0,
			H3D::Inst< H3D::MFString > _stateObjectivesField = 0
);



	virtual ~StateMachine()
	{
		cout<<"StateMachine Destroyed"<<endl;
	}

	/**
	 * Initialize the StateMachine  Node
	 */
	void initialize();

	/**
	 * Traverse the Scene Graph for StateMachine Node
	 */
	void traverseSG( H3D::TraverseInfo &ti );


	/**
	 * The H3DNodeDatabase for this node.
	 */
	static H3D::H3DNodeDatabase database;

   //TODO: should not use Vec3f for distance and force values
	auto_ptr<H3D::MFVec3f> triggerPointsRequiredField; //Point to touch in state 1
	auto_ptr<H3D::MFFloat> distancesRequiredField; //A vector with the maximum distance allowed from above points
	auto_ptr<H3D::MFFloat> forcesRequiredField; 	//A vector with the needed force on each point to get to next state
	auto_ptr<H3D::MFInt32> deviceShapesRequiredField; //Elevator or drill
	auto_ptr<H3D::MFFloat> fractionOfExpertRequiredField; //A vector with the minimum amount of drilled expert voxels user needs before going to next state
	auto_ptr<H3D::MFFloat> rotationAnglesRequiredField; //A vector with the rotation of the device required for the next step
	auto_ptr<H3D::MFString> partitionNamesField; //A vector with the partition which will be removed when the state is completed
	auto_ptr<H3D::MFInt32 > drillingStepsField; //Drillingstep of the current state

	auto_ptr<H3D::SFInt32> currentDeviceShapeField; //Current force on haptic device
	auto_ptr<H3D::SFVec3f> currentForceField; //Current force on haptic device
	auto_ptr<H3D::SFVec3f> drillPosField; //Current position of haptic device
	auto_ptr<H3D::SFVec3f> currentRotationAnglesField; //Current rotation of the haptic pen

	auto_ptr<H3D::SFInt32> stateField; //Current state
	auto_ptr<H3D::SFNode > volumeModelField; //Volume node
	auto_ptr<H3D::SFNode > drillingStepsNodeField; //Drillingsteps node

	auto_ptr<H3D::MFString > stateObjectivesField; //Objectives with each state

	/**
	 * Functions used by the GUI (forsApp).
	 */
	SFFloat getDistance();
	SFFloat getAbsoluteForce();
	void 	nextState();

private:
	VolumeModel *volumeModelNode;
	DrillingStepsNode *drillingStepsNode;
	SFFloat absoluteForce;
	SFFloat distance;
	SFFloat rotationAngle;
	SFFloat initialRotationAngle;
	int currentDrillingStep;
	unsigned int maxNrOfStates;

	void printState();
	void resetState();


	/**
	 * \brief Exception class for StateMachine Model
	 */
	class StateMachineException: public runtime_error
	{
	public:

		/**
		 * Constructor
		 */
		StateMachineException(const string& what) :
			runtime_error("Exception within StateMachine: " + what)
			{
			}
	};
};

}

#endif /* STATEMACHINE_H_ */
