#include "VolumeModel.h"

using namespace std;
using namespace FS;

// Add this node to the H3DNodeDatabase system.
H3D::H3DNodeDatabase VolumeModel::database(
		"VolumeModel",
		&(newInstance<VolumeModel>),
		typeid(VolumeModel),
		&X3DChildNode::database);

namespace VolumeModelInternals
{
H3D::FIELDDB_ELEMENT(VolumeModel, imageField,                   									INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, materialSegmentationField,        								INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, forbiddenSegmentationField,        								INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, partitionField,        											INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, expertDrillingStepsField,         								INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, timeStampsField,              									INITIALIZE_ONLY );
H3D::FIELDDB_ELEMENT(VolumeModel, dimensionsField,              									INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, materialSegmentationModelField,   								INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, forbiddenSegmentationModelField,   								INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, partitionModelField,   											INPUT_OUTPUT );
H3D::FIELDDB_ELEMENT(VolumeModel, loggingNodeField,   											INPUT_OUTPUT );
}

VolumeModel::VolumeModel(H3D::Inst<SFTexture3DNode>     imageField,
		H3D::Inst< MFUnsignedInt8FromNrrdFile >         materialSegmentationField,
		H3D::Inst< MFUnsignedInt8FromNrrdFile >         forbiddenSegmentationField,
		H3D::Inst< MFUnsignedInt8FromNrrdFile >         partitionField,
		H3D::Inst< MFUnsignedInt8FromNrrdFile >         expertDrillingStepsField,
		H3D::Inst< MFDoubleFromNrrdFile >               timeStampField,
		H3D::Inst< H3D::SFVec3f >                       dimensionsField,
		H3D::Inst< H3D::SFNode >                        materialSegmentationModelField,
		H3D::Inst< H3D::SFNode >                        forbiddenSegmentationModelField,
		H3D::Inst< H3D::SFNode >                        partitionModelField,
		H3D::Inst< H3D::SFNode >                        loggingNodeField)

:imageField(imageField),
materialSegmentationField(materialSegmentationField),
forbiddenSegmentationField(forbiddenSegmentationField),
partitionField(partitionField),
expertDrillingStepsField(expertDrillingStepsField),
timeStampsField(timeStampField),
dimensionsField(dimensionsField),
materialSegmentationModelField(materialSegmentationModelField),
forbiddenSegmentationModelField(forbiddenSegmentationModelField),
partitionModelField(partitionModelField),
loggingNodeField(loggingNodeField),
updateFrequency(5.0)

{
	cout << "Volume Model Constructing"<<endl;

	type_name = "VolumeModel";
	database.initFields( this );

	// images not loaded in a separate thread
	H3D::X3DTextureNode::load_images_in_separate_thread = false;

	cout << "Volume Model Constructed"<<endl<<endl;
}

VolumeModel::~VolumeModel()
{
	delete[] originalImageData;

	if( startTime )
	{
		delete startTime;
	}

	cout << "Volume Model Destructed"<<endl;
}

void VolumeModel::initialize()
{
	cout<<"Volume Model Initializing"<<endl;

	startTime = new H3DUtil::TimeStamp();

	// ensure that the image node has been included
	if(imageField->getValue()==NULL)
		throw VolumeModelException("Expected the imageField to be set within the x3d file");

	// make a copy of the image data so we can restore this info if needed
	H3D::Image3DTexture *texture = static_cast< H3D::Image3DTexture * >(imageField->getValue());
	imageData = (texture->image)->getValue();

	width  = imageData->width();
	height = imageData->height();
	depth  = imageData->depth();

	maxx = width -1;
	maxy = height -1;
	maxz = depth -1;

	// a cube is generated around the voxels that are removed so that when the screen is updated
	// only the voxels in this cube are updated
	imageUpdatedOnScreen = true;

	size = (unsigned long)width*height*depth;
	spacing = (imageData->pixelSize()).x;
	dimensionsField->setValue(H3D::Vec3f(width*spacing,height*spacing,depth*spacing));
	volumePerVoxel = 3*spacing;

	// save a copy of the image in its original condition
	originalImageData = new H3DUtil::RGBA[size];
	for(unsigned long i = 0; i < size; i++)
	{
		int z=i/(height*width);
		int y=(i-z*height*width)/width;
		int x=i-z*height*width-y*width;
		originalImageData[i] = imageData->getPixel(x,y,z);
	}

	// Initialize fields and check for sanity
	initializeField(materialSegmentationField.get(), MaterialSegmentation);
	initializeField(forbiddenSegmentationField.get(), ForbiddenSegmentation);
	initializeField(partitionField.get(), Partition);
	initializeField(expertDrillingStepsField.get(), ExpertDrillingStep);

	// Volume Model is dependent on Logging Node so insure first that such a child node exists
	if(loggingNodeField->getValue()==NULL)
		throw VolumeModelException ("Expected Volume Model to have a Logging Node child node");

	// validate time stamp data
	if(timeStampsField->empty())
	{
		cout<<"Initializing Time Stamp Data according to Image file dimensions"<<endl;
		timeStamps.reset(new MFDoubleFromNrrdFile());
		timeStamps->setValueFromImageData(imageData,-1);
	}
	else
	{
		timeStamps.reset(new MFDoubleFromNrrdFile(timeStampsField.get()));

		if(timeStamps->GetNoOfDimensions()!=3)
			throw VolumeModelException ("Expected the Time Stamp Data to be 3 dimensional");

		if(imageData->width()!=timeStamps->GetDimensionLengths()[0])
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same width");
		ForbiddenSegmentationModel* forbiddenSegmentationModel = static_cast< ForbiddenSegmentationModel * >(forbiddenSegmentationModelField->getValue());
		const std::vector<unsigned char> segmentIds = forbiddenSegmentationModel->getSegmentIds();

		if(imageData->height()!=timeStamps->GetDimensionLengths()[1])
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same height");

		if(imageData->depth()!=timeStamps->GetDimensionLengths()[2])
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same depth");

		if((imageData->pixelSize()).x !=(float)(timeStamps->GetSpacings()[0]))
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the x-axis");

		if((imageData->pixelSize()).y !=(float)(timeStamps->GetSpacings()[1]))
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the y-axis");

		if((imageData->pixelSize()).z !=(float)(timeStamps->GetSpacings()[2]))
			throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the z-axis");
	}

	// create a material remaining field
	cout<<"Initializing Material Remaining Data according to Segment data"<<endl;
	materialSegmentationField->GetMaskedVersion(materialRemainingField);

	global = H3D::Vec3f(0,0,0);
	relative = H3D::Vec3f(0,0,0);
	indexZeroPos = H3D::Vec3f(0-(maxx/2.0f)*spacing, 0-(maxy/2.0f)*spacing, 0-(maxz/2.0f)*spacing); // Volume centered around 0,0,0

	// Create a multimap vector from the timestamp-file. This will make it much faster to do volume-
	// playback since it will not be necessary to traverse the whole volume for each new image-update.
	for(unsigned int i =0;i<width;i++)
		for(unsigned int j =0;j<height;j++)
			for(unsigned int k =0;k<depth;k++)
			{
				Vec3i vec(i,j,k);
				double timestamp = getTimeStamp(vec);
				if(timestamp!=-1)
					timestamp_map.insert(pair<double,Vec3i>(timestamp, vec));
			}

	X3DChildNode::initialize();

	cout<<"Volume Model Initialized"<<endl<<endl;
}

void VolumeModel::initializeField(MFUnsignedInt8FromNrrdFile* field, FieldId fieldId)
{
		string errorId;
		MaterialSegmentationModel* materialSegmentationModel;
		std::vector<unsigned char> segmentIds;
		ForbiddenSegmentationModel* forbiddenSegmentationModel;
		PartitionModel* partitionModel;

		switch(fieldId)
		{
			case MaterialSegmentation:
				errorId = "Material Segmentation Field";
				materialSegmentationModel = static_cast< MaterialSegmentationModel * >(materialSegmentationModelField->getValue());
				segmentIds = materialSegmentationModel->getSegmentIds();
				break;

			case Partition:
				errorId = "Partition Field";
				partitionModel = static_cast< PartitionModel * >(partitionModelField->getValue());
				segmentIds = partitionModel->getPartitionIds();
				break;

			case ExpertDrillingStep:
				errorId = "Expert Drilling Step Field";
				break;

			case ForbiddenSegmentation:
				errorId = "Forbidden Segmentation Field";
				forbiddenSegmentationModel = static_cast< ForbiddenSegmentationModel * >(forbiddenSegmentationModelField->getValue());
				segmentIds = forbiddenSegmentationModel->getSegmentIds();
				break;

			default:
				throw VolumeModelException ("Unknown FieldId in initializing of Field");
		}

		if(field->empty()) 	// material segmentation field has not been initialized from a nrrd file
		{
			cout<<"Initializing " + errorId + " Data according to Image file dimensions.";
			cout<<"All voxels given a segment Id of: "<< AIR_SEGMENT_ID;
			field->setValueFromImageData(imageData,AIR_SEGMENT_ID);
		}
		else				          			// segment field has been initialized from a nrrd file
		{
			if(field->GetNoOfDimensions()!=3)
				throw VolumeModelException ("Expected the " + errorId + " Data to be 3 dimensional");

			if(imageData->width()!=field->GetDimensionLengths()[0])
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same width");

			if(imageData->height()!=field->GetDimensionLengths()[1])
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same height");

			if(imageData->depth()!=field->GetDimensionLengths()[2])
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same depth");

			if((imageData->pixelSize()).x !=(float)(field->GetSpacings()[0]))
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same spacing in the x-axis");

			if((imageData->pixelSize()).y !=(float)(field->GetSpacings()[1]))
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same spacing in the y-axis");

			if((imageData->pixelSize()).z !=(float)(field->GetSpacings()[2]))
				throw VolumeModelException ("Expected Image Data and " + errorId + " Data to have the same spacing in the z-axis");
		}

			if(fieldId != ExpertDrillingStep)
			{
				// loop through all voxels
				for(unsigned int i =0;i<width;i++)
					for(unsigned int j =0;j<height;j++)
						for(unsigned int k =0;k<depth;k++)
						{
							Vec3i vec(i,j,k);
							unsigned char segmentId = field->getValue()[getIndex(vec)];
							std::vector<unsigned char>::const_iterator ii;

							bool foundSegmentId = false;

							// loop through the segment Ids in Segment Model
							for(ii=segmentIds.begin();ii<segmentIds.end();ii++)
								if(segmentId == *ii)
								{
									foundSegmentId = true;
									break;
								}

							if(!foundSegmentId)
							{
								stringstream failStringStream;
								failStringStream << "The Segment Id: " << (unsigned int)segmentId << " at voxel: ("<<i<<","<<j<<","<<k<<") does not exist within Material Segmentation Model";
								throw VolumeModelException(failStringStream);
							}
						}
			}
		}


void VolumeModel::traverseSG(H3D::TraverseInfo &ti)
{
	// update the region on screen where we have bored
	H3DUtil::TimeStamp now = H3DUtil::TimeStamp::now();
	if((double)now > ((double)lastUpdateTime + 1/updateFrequency))
	{
		lastUpdateTime = now;
		if(imageUpdatedOnScreen!=true)
		{
			imageField->getValue()->image->endEditing();
			imageUpdatedOnScreen = true;
		}
	}
}

void VolumeModel::setTimeStampFile(const string& nrrdFilePath )
{
	cout << "Setting Timestamp Data from file: " << nrrdFilePath <<endl;

	timeStamps->setValueFromString(nrrdFilePath);

	if(timeStamps->GetNoOfDimensions()!=3)
		throw VolumeModelException ("Expected the Time Stamp Data to be 3 dimensional");

	if(imageData->width()!=timeStamps->GetDimensionLengths()[0])
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same width");

	if(imageData->height()!=timeStamps->GetDimensionLengths()[1])
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same height");

	if(imageData->depth()!=timeStamps->GetDimensionLengths()[2])
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same depth");

	if((imageData->pixelSize()).x !=(float)(timeStamps->GetSpacings()[0]))
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the x-axis");

	if((imageData->pixelSize()).y !=(float)(timeStamps->GetSpacings()[1]))
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the y-axis");

	if((imageData->pixelSize()).z !=(float)(timeStamps->GetSpacings()[2]))
		throw VolumeModelException ("Expected Image Data and Time Stamp Data to have the same spacing in the z-axis");

	timestamp_map.clear();
	// Create a multimap vector from the timestamp-file. This will make it much faster to do volume-
	// playback since it will not be necessary to traverse whole volume for each new image-update.
	for(unsigned int i =0;i<width;i++)
		for(unsigned int j =0;j<height;j++)
			for(unsigned int k =0;k<depth;k++)
			{
				Vec3i vec(i,j,k);
				double timestamp = getTimeStamp(vec);
				if(timestamp!=-1)
					timestamp_map.insert(pair<double,Vec3i>(timestamp, vec));
			}
}


double VolumeModel::getElapsedTime()
{
	H3DUtil::TimeStamp currentTime;
	return currentTime - *startTime;
}

void VolumeModel::resetStartTime()
{
	//Starttime is set to current time when drilling has started, see ADrillableNode
	if( startTime )
		delete startTime;

	startTime = new H3DUtil::TimeStamp();
}

unsigned int VolumeModel::getHeight()
{
	return height;
}

unsigned int VolumeModel::getWidth()
{
	return width;
}

unsigned int VolumeModel::getDepth()
{
	return depth;
}

float VolumeModel::getSpacing()
{
	return spacing;
}

unsigned int VolumeModel::getIndex(Vec3i p)
{
	return ((unsigned int)p.z * height * width + (maxy-(unsigned int)p.y) * width + (unsigned int)p.x);
}

/****************************************************************/
void VolumeModel::setImageValue(Vec3i p, float r, float g, float b)
{
	if(imageUpdatedOnScreen==true)
	{
		imageField->getValue()->image->beginEditing();
		imageUpdatedOnScreen = false;
	}
	imageField->getValue()->image->setPixel(p.x,maxy-p.y,p.z,H3D::RGBA(r,g,b,1));
}

H3D::H3DFloat VolumeModel::getImageValue(Vec3i p)
{
	return imageData->getPixel(p.x,maxy-p.y,p.z).r;
}

H3D::H3DFloat VolumeModel::getImageValue(unsigned int i)
{
	int z=i/(height*width);
	int y=(i-z*height*width)/width;
	int x=i-z*height*width-y*width;
	return imageData->getPixel(x,maxy-y,z).r;
}

H3D::H3DFloat VolumeModel::getOriginalImageValue(Vec3i p)
{
	return originalImageData[getIndex(p)].r;
}

/****************************************************************/

void VolumeModel::setMaterialSegmentationId(Vec3i p, unsigned char value)
{
	materialSegmentationField->setValue(getIndex(p),value);
}

unsigned char VolumeModel::getMaterialSegmentationId(Vec3i p)
{
	return materialSegmentationField->getValue()[getIndex(p)];
}

unsigned char VolumeModel::getMaterialSegmentationId(unsigned int index)
{
	return materialSegmentationField->getValue()[index];
}

/****************************************************************/

unsigned char VolumeModel::getForbiddenSegmentationId(Vec3i p)
{
	return forbiddenSegmentationField->getValue()[getIndex(p)];
}

void VolumeModel::incrementNoOfVoxelsBoredByUserForForbiddenSegment(Vec3i p)
{
	ForbiddenSegmentationModel* forbiddenSegmentationModel = static_cast< ForbiddenSegmentationModel * >(forbiddenSegmentationModelField->getValue());
	forbiddenSegmentationModel->incrementNoOfVoxelsBoredByUser(getForbiddenSegmentationId(p));
}

void VolumeModel::incrementNoOfVoxelsBoredByUserForSegment(Vec3i p)
{
	MaterialSegmentationModel* materialSegmentationModel = static_cast< MaterialSegmentationModel * >(materialSegmentationModelField->getValue());
	materialSegmentationModel->incrementNoOfVoxelsBoredByUser(getMaterialSegmentationId(p));
}

const vector<Vec3i> VolumeModel::getAllForbiddenVoxels()
{
	vector<Vec3i> forbiddenVoxels;

	// loop through all voxels
	for(unsigned int i =0;i<width;i++)
		for(unsigned int j =0;j<height;j++)
			for(unsigned int k =0;k<depth;k++)
			{
				Vec3i vec(i,j,k);
				if(getForbiddenSegmentationId(vec)!=UNDEFINED_FORBIDDEN_SEGMENT_ID)
					forbiddenVoxels.push_back(vec);
			}

	std::sort(forbiddenVoxels.begin(),forbiddenVoxels.end());

	return forbiddenVoxels;
}

/****************************************************************/

void VolumeModel::setPartitionId(Vec3i p, unsigned char value)
{
	partitionField->setValue(getIndex(p),value);
}

unsigned char VolumeModel::getPartitionId(Vec3i p)
{
	return partitionField->getValue()[getIndex(p)];
}

unsigned char VolumeModel::getPartitionId(unsigned int index)
{
	return partitionField->getValue()[index];
}

/****************************************************************/

bool VolumeModel::isExpertDrillingStep(Vec3i p, unsigned char expertDrillingStepId)
{
	return expertDrillingStepsField->getValue()[getIndex(p)]==expertDrillingStepId?true:false;
}

unsigned char VolumeModel::getExpertDrillingStepId(Vec3i p)
{
	return expertDrillingStepsField->getValue()[getIndex(p)];
}

void VolumeModel::setExpertDrillingStepId(Vec3i p, unsigned char drillingStepId)
{
	return expertDrillingStepsField->setValue(getIndex(p), drillingStepId);
}

bool VolumeModel::isDrilledByUser(Vec3i p)
{
	if(materialRemainingField.getValue()[getIndex(p)]==0.0 && materialSegmentationField->getValue()[getIndex(p)]!=AIR_SEGMENT_ID)
		return true;
	return false;
}

void VolumeModel::printUniqueDrilledBy()
{
	vector<unsigned char> list;

	for(unsigned int i =0;i<getWidth();i++)
		for(unsigned int j =0;j<getHeight();j++)
			for(unsigned int k =0;k<getDepth();k++)
			{
				Vec3i vec(i,j,k);
				vector<unsigned char>::iterator iter;
				unsigned char value = expertDrillingStepsField->getValue()[getIndex(vec)];
				bool found = false;
				for(iter=list.begin(); iter<list.end(); iter++)
					if(*iter == value)
					{

						found = true;
						break;
					}
				if(!found)
				{
					list.push_back(value);
					cout << hex <<(unsigned int)value << endl;
				}
			}
}

/****************************************************************/

void VolumeModel::setMaterialRemaining(Vec3i p, float value)
{
	materialRemainingField.setValue(getIndex(p),value);
}

float VolumeModel::getMaterialRemaining(Vec3i p)
{
	return materialRemainingField.getValue()[getIndex(p)];
}

float VolumeModel::getMaterialRemaining(unsigned int index)
{
	return materialRemainingField.getValue()[index];
}
/****************************************************************/

void VolumeModel::setTimeStamp(Vec3i p, double value)
{
	timeStamps->setValue(getIndex(p),value);
	timestamp_map.insert(pair<double,Vec3i>(value, p));
}

double VolumeModel::getTimeStamp(Vec3i p)
{
	return timeStamps->getValue()[getIndex(p)];
}

double VolumeModel::getTimeStamp(unsigned int index)
{
	return timeStamps->getValue()[index];
}

string VolumeModel::saveTimeStamps()
{
	cout << "VolumeModel: Saving Timestamps" << endl;

	unsigned int noOfDimensions = 3;

	vector<unsigned int> dimensionLengths;
	dimensionLengths.push_back(width);
	dimensionLengths.push_back(height);
	dimensionLengths.push_back(depth);

	vector<double> spacings;
	spacings.push_back(spacing);
	spacings.push_back(spacing);
	spacings.push_back(spacing);

	LoggingNode* loggingNode = static_cast< LoggingNode * >(loggingNodeField->getValue() );
	string userLoggingPath = loggingNode->userLoggingPathField->getValue();

	// construct string with current date and time
	time_t rawTime = time(0);
	struct tm* timeInfo;
	char strBuffer [80];
	timeInfo = localtime ( &rawTime );
	strftime (strBuffer,80,"%y%m%d%H%M%S",timeInfo);
	string timeString(strBuffer);

	// construct file name
	string filePath = userLoggingPath+"/";
	filePath.append(timeString).append(".nrrd");

	cout << "to file named: " << filePath <<endl;

	timeStamps->saveDataToNrrdFile(filePath,noOfDimensions,dimensionLengths,spacings);

	return filePath;
}

void VolumeModel::saveStatistics()
{

	LoggingNode* loggingNode = static_cast< LoggingNode * >(loggingNodeField->getValue() );
	string userLoggingPath = loggingNode->userLoggingPathField->getValue();

	MaterialSegmentationModel* materialSegmentationNode =
		static_cast< MaterialSegmentationModel * >(materialSegmentationModelField->getValue());

    const float mm3PerM3 = 1000000.0;
	// construct string with current date and time
	time_t rawTime = time(0);
	struct tm* timeInfo;
	char strBuffer [80];
	timeInfo = localtime ( &rawTime );
	strftime (strBuffer,80,"%y%m%d%H%M%S",timeInfo);
	string timeString(strBuffer);

	// construct file name
	string filePath = userLoggingPath+"/";
	filePath.append(timeString).append("_stat").append(".txt");

	//creating file
	ofstream file;
	file.open(filePath.c_str(),ios::out);

	/*Data*/
	//Total numbers of voxels and segments
	int totVoxels = height*width*depth;
	file<<"Total numbers of voxels: " << totVoxels << " (including Air)" << endl;
	file << "Number of Segments: "<<materialSegmentationNode->getNoOfSegments()<<endl;

	//Total time on case
	double timeSpendBoring = 0;
	if(timestamp_map.size()>1)
		timeSpendBoring = (*(timestamp_map.rbegin())).first - (*(timestamp_map.begin())).first;
	file << "Time spent boring: "<<timeSpendBoring<<" secs"<<endl<<endl;

	//Number of Forbidden voxels
	ForbiddenSegmentationModel* forbiddenSegmentationModel = static_cast< ForbiddenSegmentationModel * >(forbiddenSegmentationModelField->getValue());

	//Number of voxels drilled in each segment
	int noOfSegments = materialSegmentationNode->getNoOfSegments();
	int voxelsEachSegArray[noOfSegments];

	for (int i = 0; i < noOfSegments; i++)
	{
		voxelsEachSegArray[i] = 0;
	}

	multimap<double, Vec3i>::iterator it;
	for ( it=timestamp_map.begin() ; it !=timestamp_map.end(); it++ )
	{
		int temp = getMaterialSegmentationId((*it).second);
		voxelsEachSegArray[temp] = voxelsEachSegArray[temp] + 1;
	}
	file << " Voxels drilled in " << endl;
	for (int i = 1; i < noOfSegments; i++) //Start at 1 since 0=Air
	{
		file << "Segment " << i << " - " << materialSegmentationNode->getName(i) << ": " << voxelsEachSegArray[i] << endl;
	}

	//Forbidden segments
	file << endl;
	file << "(TODO These values are not the same as displayed in GUI. Check code...) Number of voxels drilled in forbidden areas" <<endl;
	file << "Wisdom Backside: " << forbiddenSegmentationModel->noOfVoxelsBoredByUser->getValue()[0] << endl;
	file << "Other teeth: " << forbiddenSegmentationModel->noOfVoxelsBoredByUser->getValue()[1] <<endl<<endl;

	file.close();
	cout << "Saving statistics to file " << filePath << endl;
}

 unsigned int VolumeModel::noOfVoxelsRemoved(unsigned int segmentId){

		//Number of voxels drilled in each segment
		int noOfRemovedVoxels = 0;


		multimap<double, Vec3i>::iterator it;
		for ( it=timestamp_map.begin() ; it !=timestamp_map.end(); it++ )
		{
			if (segmentId == getMaterialSegmentationId((*it).second))
				noOfRemovedVoxels++;
		}
		return noOfRemovedVoxels;
 }

/****************************************************************/

H3D::Vec3f VolumeModel::getGlobalCoordinate()
{
	return global;
}

H3D::Vec3f VolumeModel::getGlobalCoordinate(int i)
{
	return getLocalCoordinate(i)+global;
}
H3D::Vec3f VolumeModel::getGlobalCoordinate(Vec3i p)
{
	return getLocalCoordinate(p)+global;
}

H3D::Vec3f VolumeModel::getLocalCoordinate()
{
	return H3D::Vec3f(0,0,0);
}
H3D::Vec3f VolumeModel::getLocalCoordinate(int i)
{
	Vec3i p;
	p.z=i/(height*width);
	p.y=(i-p.z*height*width)/width;
	p.x=i-p.z*height*width-p.y*width;
	return(getLocalCoordinate(p));
}
H3D::Vec3f VolumeModel::getLocalCoordinate(Vec3i p)
{
	return H3D::Vec3f
	(p.x*spacing,p.y*spacing,p.z*spacing)+indexZeroPos;
}

H3D::Vec3f VolumeModel::getRelativeCoordinate()
{
	return getGlobalCoordinate()-relative;
}
H3D::Vec3f VolumeModel::getRelativeCoordinate(int i)
{
	return getGlobalCoordinate(i)-relative;
}
H3D::Vec3f VolumeModel::getRelativeCoordinate(Vec3i p)
{
	return getGlobalCoordinate(p)-relative;
}

// The coordinates of the center of our object, in GLOBAL coordinates
void VolumeModel::setGlobalCoordinate(H3D::Vec3f c)
{
	global = c;
}

// The coordinates of relative (reference) point, in GLOBAL coordinates
void VolumeModel::setRelativeCoordinate(H3D::Vec3f c)
{
	relative = c;
}

Vec3i VolumeModel::getIndexLocal(H3D::Vec3f c)
{
	Vec3i i;
	H3D::Vec3f voxel=(c-indexZeroPos)/spacing;
	i.x = (int)voxel.x;
	i.y = (int)voxel.y;
	i.z = (int)voxel.z;
	if(i.x<0 || i.y<0 || i.z<0 || (unsigned int)i.x>maxx || (unsigned int)i.y>maxy || (unsigned int)i.z>maxz){
		i.x=-1;
		i.y=-1;
		i.z=-1;
	}
	return i;
}
Vec3i VolumeModel::getIndexLocalWall(H3D::Vec3f c)
{
	Vec3i i;
	H3D::Vec3f voxel=(c-indexZeroPos)/spacing;
	i.x = (int)voxel.x;
	i.y = (int)voxel.y;
	i.z = (int)voxel.z;
	return wall(i);
}
Vec3i VolumeModel::getIndexGlobal(H3D::Vec3f c)
{
	return getIndexLocal(c-global);
}
Vec3i VolumeModel::getIndexGlobalWall(H3D::Vec3f c)
{
	return getIndexLocalWall(c-global);
}
Vec3i VolumeModel::getIndexRelative(H3D::Vec3f c)
{
	return getIndexLocal(c-relative);
}
Vec3i VolumeModel::getIndexRelativeWall(H3D::Vec3f c)
{
	return getIndexLocalWall(c-relative);
}

// For safe in-boundary calculations, returns 0 if coordinate < 0, max if
// coordinate above max
Vec3i VolumeModel::wall(Vec3i i)
{
	if(i.x<0) i.x=0;
	if(i.y<0) i.y=0;
	if(i.z<0) i.z=0;
	if((unsigned int)i.x>maxx) i.x=maxx;
	if((unsigned int)i.y>maxy) i.y=maxy;
	if((unsigned int)i.z>maxz) i.z=maxz;
	return i;
}

// Returns a list of 6 cartesian neighbours of given scalar id. If no neighbour
// exists (out of border), id will be set to -1.
void VolumeModel::getNeighbours(Vec3i *vector, Vec3i id)
{
	vector[0]=Vec3i(id.x-1,id.y,id.z);
	if(vector[0].x<0)
		vector[0] = Vec3i(-1,-1,-1);

	vector[1]=Vec3i(id.x+1,id.y,id.z);
	if((unsigned int)vector[1].x>maxx)
		vector[1] = Vec3i(-1,-1,-1);

	vector[2]=Vec3i(id.x,id.y-1,id.z);
	if(vector[2].y<0)
		vector[2] = Vec3i(-1,-1,-1);

	vector[3]=Vec3i(id.x,id.y+1,id.z);
	if((unsigned int)vector[3].y>maxy)
		vector[3] = Vec3i(-1,-1,-1);

	vector[4]=Vec3i(id.x,id.y,id.z-1);
	if(vector[4].z<0)

		vector[5]=Vec3i(id.x,id.y,id.z+1);
	if((unsigned int)vector[5].z>maxz)
		vector[5] = Vec3i(-1,-1,-1);
}

bool VolumeModel::isInsideVolume( H3D::Vec3f c )
{
	H3D::Vec3f max_v( spacing * (maxx+1) / 2,
			spacing * (maxy+1) / 2,
			spacing * (maxz+1) / 2 );

	return (c.x >= -max_v.x && c.x <= max_v.x &&
			c.y >= -max_v.y && c.y <= max_v.y &&
			c.z >= -max_v.z && c.z <= max_v.z );
}

float VolumeModel::getVolumeFromVoxels(int noOfVoxels)
{
	return volumePerVoxel*noOfVoxels;
}

void VolumeModel::removePartitionByName(string partitionName)
{
	PartitionModel* partitionModel = static_cast< PartitionModel * >(partitionModelField->getValue());
	unsigned char idOfPartitionToRemove = partitionModel->getId(partitionName);

	double elapsedTime = getElapsedTime();
	for(unsigned int i =0;i<width;i++)
		for(unsigned int j =0;j<height;j++)
			for(unsigned int k =0;k<depth;k++)
			{
				Vec3i vec(i,j,k);
				if(getPartitionId(vec)==idOfPartitionToRemove)
				{
					setMaterialRemaining(vec,0.0);
					setImageValue(vec,0,0,0);
					if(-1==getTimeStamp(vec))
					{
						//The voxel has previously not been removed already
						setTimeStamp(vec,elapsedTime);
					}
					//setAsCurrentExpertDrillingStep(vec);
						//setThatUserHasBoredThisDrillingStep(vec,elapsedTime);
				}
			}
}
